package FileFormatting;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;

public class GeneGangPairs extends LoadData{

	//Fields
	public static String QuerySetsDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/JCE_Files/QuerySets";
	public static String PBCV1ext = "/PBCV1_Base";

	public static HashSet<Integer[]> LocalMaxPairs;
	
	public static LinkedList<String> Organisms = new LinkedList<String>();
	public static LinkedHashMap<LinkedList<String>, Integer> BumpCounts = new LinkedHashMap<LinkedList<String>, Integer>();
	public static HashSet<Integer> AllClusters = new HashSet<Integer>();
	
	//Class
	public static class MatchCollection{
		public String OrgName;
		public int ClusterA;
		public int ClusterB;
		public LinkedList<GenomicFeature> ListA;
		public LinkedList<GenomicFeature> ListB;
		public LinkedList<Integer> Distances;
	}
	
	//Methods
	
	//Determine distances, as strings for eventual export
	public static LinkedList<String> DistRange(int MinDist, int MaxDist, LinkedList<GenomicFeature> Features){
		
		//store entries in hash set
		HashSet<String> HS = new HashSet<String>();
		
		String str = "";
		//parse through
		for (int i = 0; i < Features.size(); i++){
			for (int j = 0; j < Features.size(); j++){
				if (i != j){
					
					//get features
					GenomicFeature F1 = Features.get(i);
					GenomicFeature F2 = Features.get(j);
					
					//only proceed with comparison for appropriate genomes
					//if (F1.ClusterNum != 0 && F2.ClusterNum != 0){
					
					//only proceed with comparison for appropriate genomes
					if (F1.ClusterNum != 0 && F2.ClusterNum != 0 &&
							!LargeClusters.contains(F1.ClusterNum) &&
							!LargeClusters.contains(F2.ClusterNum)){	
						
						//if the genes are far enough away
						if ((F2.CenterPos - F1.CenterPos > MinDist &&	//Case: F1 < F2
								F2.CenterPos - F1.CenterPos < MaxDist)||
								(F1.CenterPos - F2.CenterPos > MinDist &&//Case: F1 > F2
								F1.CenterPos - F2.CenterPos < MaxDist)){
							
							//sort clusters
							Integer[] Clusters = new Integer[2];
							if (F1.ClusterNum <= F2.ClusterNum){
								Clusters[0] = F1.ClusterNum;
								Clusters[1] = F2.ClusterNum;
							} else {
								Clusters[0] = F2.ClusterNum;
								Clusters[1] = F1.ClusterNum;
							}
							
							//parse to string
							str = String.valueOf(Clusters[0]) 
									+ "; " + String.valueOf(Clusters[1]);
							
							//add to hash set
							HS.add(str);

						}
					}
				}
			}
		}
		
		//output - generate linked list from has hset
		LinkedList<String> OutputList = new LinkedList<String>(HS);
		
		//return list
		return OutputList;
	}
	
	//Determine distances, as internal for processing
	public static LinkedList<Integer[]> DistRangeAsArrays(int MinDist, int MaxDist, String GenomeName){
		
		//store entries in hash set
		HashSet<Integer[]> HS = new HashSet<Integer[]>();
		
		//Retrieve featuers
		LinkedList<GenomicFeature> Features = Chloroviruses.get(GenomeName);
		
		//parse through
		for (int i = 0; i < Features.size(); i++){
			for (int j = 0; j < Features.size(); j++){
				if (i != j){
					
					//get features
					GenomicFeature F1 = Features.get(i);
					GenomicFeature F2 = Features.get(j);
					
					//only proceed with comparison for appropriate genes
					//genes must be core clusters, and must not be in the set of LargeClusters.
					if (F1.ClusterNum != 0 && F2.ClusterNum != 0 &&
							!LargeClusters.contains(F1.ClusterNum) &&
							!LargeClusters.contains(F2.ClusterNum)){
					
						//if the genes are far enough away
						if ((F2.CenterPos - F1.CenterPos > MinDist &&	//Case: F1 < F2
								F2.CenterPos - F1.CenterPos < MaxDist)||
								(F1.CenterPos - F2.CenterPos > MinDist &&//Case: F1 > F2
								F1.CenterPos - F2.CenterPos < MaxDist)){
							
							//sort clusters
							Integer[] Clusters = new Integer[2];
							if (F1.ClusterNum <= F2.ClusterNum){
								Clusters[0] = F1.ClusterNum;
								Clusters[1] = F2.ClusterNum;
							} else {
								Clusters[0] = F2.ClusterNum;
								Clusters[1] = F1.ClusterNum;
							}
							
							//add to hash set
							HS.add(Clusters);

						}
					}
				}
			}
		}
		
		//output - generate linked list from has hset
		LinkedList<Integer[]> OutputList = new LinkedList<Integer[]>(HS);
		
		//return list
		return OutputList;
	}
	
	//Process a query set, return fraction of gene gangs
	public static double FractionGeneGangPairs(LinkedList<Integer[]> Pairs, String GenomeName, int DistChangeThreshold, double ConservationStringency){
		
		//Initialize output
		int GeneGangPairs = 0;
		
		//determine matches for each pair
		for (Integer[] x : Pairs){
			
			//counter for individual clusters
			int MCMatches = 0;
			
			//Generate an appropriate match collection
			MatchCollection Source = GenerateMatchCollection(x, GenomeName);
			
			//How often does this match occur in other genomes?
			MCMatches = MatchesInOtherGenomes(Source, DistChangeThreshold);
			
			//evaluate if enough matched to count this set as a conserved gene gang pair.
			double FracConserved = (double) ((double) (MCMatches) / (double) (Chloroviruses.keySet().size() - 1));
			
			//breakpoint
			if (MCMatches == 20){			
				//LocalMaxPairs.add(x);
				AllClusters.add(x[0]);
				AllClusters.add(x[1]);
			}

			//Increment counter
			if (FracConserved >= ConservationStringency){
				GeneGangPairs++;
				
//				//increment hashmap
//				if (BumpCounts.get(Organisms) != null){
//					int Val = BumpCounts.get(Organisms);
//					BumpCounts.put(Organisms,(Val+1));
//				} else {
//					BumpCounts.put(Organisms,1);
//				}
				
//				//Investigate 30 kB regions - ignore multiple copies here, just for now
//				if (Source.Distances.size() == 1){
//					
//					//store the pair
//					LocalMaxPairs.add(x);
//				}

			}
			
		}
		
		//return fraction of the original pairs that remain conserved in every case
		return (double) ((double) GeneGangPairs / (double) Pairs.size());
	}
	
	//Determine number of matches in the other genomes
	public static int MatchesInOtherGenomes(MatchCollection M, int DistChangeThreshold){
		
		/*
		 * (1) Retrieve parameters from original match collection
		 * (2) Find Matches in other genomes
		 * (3) Compare MatchSet in other genomes with the source genome
		 * (4) if the two match sets contain a plausible match, inc. counter
		 * (5) return total number of matches
		 * 
		 */
		
		//initialize list of organisms
		Organisms = new LinkedList<String>();
		
		//Initialize output
		int Matches = 0;
		
		//retrieve original clusters from supplied match collection
		Integer[] Clusters = new Integer[2];
		Clusters[0] = M.ClusterA;
		Clusters[1] = M.ClusterB;
		
		//check every match in the set
		for (String s : Chloroviruses.keySet()){
			
			//generate an analogous match collection for every case, except the source, + compare
			if (!s.equals(M.OrgName)){
				
				//generate a match collection for this set of clusters
				MatchCollection ScanCheck = GenerateMatchCollection(Clusters, s);
				
				//compare this match collection to the master
				boolean HasMatch = DistanceMatch(M, ScanCheck, DistChangeThreshold);
				
				//if it's a match, increment counter
				if (HasMatch){
					Matches++;
					Organisms.add(ScanCheck.OrgName);
				}
			}
		}

		//return output
		return Matches;
	}
	
	//Perform comparison of two matches
	public static boolean DistanceMatch(MatchCollection Master, MatchCollection Slave, int DistChangeThreshold){
		
		//initialize output
		boolean ConservedDistance = false;
		int ConsDistanceCounter = 0;
		
//		//Ignore cases with multiple instances
//		if (Master.Distances.size() == 1 && Slave.Distances.size() == 1){
//			
//			for (int D1 : Master.Distances){
//				for (int D2 : Slave.Distances){
//					if (Math.abs(D1-D2) < DistChangeThreshold){
//						ConservedDistance = true;
//						ConsDistanceCounter++;
//						break;
//					}
//				}
//			}
//			
//		}
		
		
		//if even one distance is true, the whole set counts as true.
		for (int D1 : Master.Distances){
			for (int D2 : Slave.Distances){
				if (Math.abs(D1-D2) < DistChangeThreshold){
					ConsDistanceCounter++;
					ConservedDistance = true;
					break;
				}
			}
		}
		
//		//Break when disagreeement
//		if (ConservedDistance && ConsDistanceCounter != 2){
//			System.out.println("Breakpoint!");
//		}
		
		//return output
		return ConservedDistance;
	}
	
	//Generate match collection from a set of clusters and a genome.
	public static MatchCollection GenerateMatchCollection(Integer[] Clusters, String Genome){
		
		//initalize match collection, and write parameter values
		MatchCollection M = new MatchCollection();
		M.OrgName = Genome;
		M.ClusterA = Clusters[0];
		M.ClusterB = Clusters[1];
		M.ListA = new LinkedList<GenomicFeature>();
		M.ListB = new LinkedList<GenomicFeature>();
		M.Distances = new LinkedList<Integer>();
		
		//retrieve all appropriate features
		LinkedList<GenomicFeature> Features = Chloroviruses.get(Genome);
		for (GenomicFeature F : Features){
			if (F.ClusterNum == M.ClusterA){
				M.ListA.add(F);
			} else if (F.ClusterNum == M.ClusterB){
				M.ListB.add(F);
			}
		}
		
		//build distances
		for (GenomicFeature F1 : M.ListA){
			for (GenomicFeature F2 : M.ListB){
				M.Distances.add(Math.abs(F1.CenterPos - F2.CenterPos));
			}
		}
		
		//return match collection
		return M;
	}
	
	//Run through whole analysis
	public static LinkedList<Double> GeneGangPairsAnalysis(int BinSize, int NumBins, String GenomeName, int DistChangeThreshold, double ConservationStringency){
		
		//Initialize output
		LinkedList<Double> ValueList = new LinkedList<Double>();
		
		//Iterate through starting distance values
		for (int i = 0; i < (NumBins-1); i++){
		
		//bin 6: 30K - 35K
		//for (int i = 6; i <= 6; i++){
			
			//Define lower bound and upper bound for center-to-center distance
			int LowDist = BinSize*i;
			int HighDist = BinSize*(i+1);
			
			//Determine all gene pairs in the organism of interest separated by the distances specified.
			LinkedList<Integer[]> Pairs = DistRangeAsArrays(LowDist,HighDist,GenomeName);
			
			//debugging - print number of pairs, per genome
			//System.out.println(GenomeName + "\t" + LowDist + "\t" + Pairs.size());
			
			//Assumption: if even one set of (A,B) match the original, then the set is conserved.
			//Determine % of conserved gene gangs in the chloroviruses for set of gene pairs provided.
			//Gene gang strictness is specified in 'ConservationStringency'.
			double FractionConserved = FractionGeneGangPairs(Pairs, GenomeName, DistChangeThreshold, ConservationStringency);
			
			//add to list
			ValueList.add(FractionConserved);
			
			//output
			//System.out.println("From " + LowDist + " to " + HighDist + ": " + FractionConserved + " Conserved.");
			//System.out.println(LowDist + "\t" + FractionConserved);
		}
		
		//return list
		return ValueList;
	}
	
	// ====== Export ====== //
	
	//Export Range Files
	public static void ExportFilebyLines(LinkedList<String> Lines4Export, String OutputFile){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(OutputFile));
			for (String s : Lines4Export){
				bw.write(s);
				bw.write("\n");
				bw.flush();
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export suspicious local max pairs
	public static void ExportLocalMaxPairs(String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			for (Integer[] x : LocalMaxPairs){
				String str = String.valueOf(x[0]) + "; " + String.valueOf(x[1]) + "\n";
				bw.write(str);
				bw.flush();
			}
			bw.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	//main method
	public static void main(String[] args) {
		
		// ======= Initializations (from old method) ========= //
		
		// initialize fields
		LoadAllData();
		
		//pairs in local max
		LocalMaxPairs = new HashSet<Integer[]>();
		
		// ======= New Methods: gene gang pair analysis ========= //
		
		//all parameters
		int BinSize = 5000;
		int NumBins = 21;
		int DistChangeThreshold = 10000;		//10K, Phil, 10K!
		double ConservationStringency = 1.00;	//Need at least this ratio conserved to call the pair a gang.
		
		//genome-wide averages
		LinkedList<LinkedList<LinkedList<Double>>> ByGenome = new LinkedList<LinkedList<LinkedList<Double>>>();
		
		LinkedList<String> SmallerSet = new LinkedList<String>();
		SmallerSet.add("PBCV1");
		SmallerSet.add("ATCV1");
		
		String Header = ".\t" ;
		boolean WriteHeader = true;
		
		//create a set of lists for each genome
		for (String GenomeName : Chloroviruses.keySet()){
		//for (String GenomeName: SmallerSet){	
			//manage runs
			LinkedList<LinkedList<Double>> RunLists = new LinkedList<LinkedList<Double>>();
			
			//Needs to be ints to avoid floating-point arithmetic errors
			for (int i = 0; i <= 40; i++){
				
				//turn integer to double
				double C = 0.025* (double) i;
				
				//remove floating-point arithmetic errors
				C = (double)Math.round(C * 1000) / 1000;

				if (WriteHeader){
					//increment header
					Header = Header + String.valueOf(C) + "\t";
				}
				
				//Run analysis, with parameters
				 LinkedList<Double> SingleRun = GeneGangPairsAnalysis(BinSize,NumBins,GenomeName,DistChangeThreshold,C);

				 //add to set of lists
				 RunLists.add(SingleRun);
			}
			
			//once written once, header is done
			WriteHeader = false;
			
			//Add this organisms total to the list
			ByGenome.add(RunLists);
			
		}
		
		//Compress to averages
		LinkedList<LinkedList<Double>> RunLists = ByGenome.get(0);

			for (int j = 0; j < ByGenome.get(0).size(); j++){
				for (int k = 0; k < ByGenome.get(0).get(j).size(); k++){
					
					//tabulate, average by genome
					double Total = 0.0;
					for (int i = 0; i < ByGenome.size(); i++){
						Total = Total + ByGenome.get(i).get(j).get(k);
					}
					
					double Avg = Total / ((double) Chloroviruses.keySet().size());
					RunLists.get(j).set(k, Avg);
				}
			}

		//output
		//print header	
		System.out.println(Header);	
			
		//print results
		for (int i = 0; i < RunLists.get(0).size(); i++){
			
			//bin size
			int LowDist = BinSize*i;
			
			//build line
			String Line = String.valueOf(LowDist) + "\t";
			
			//add computed values
			for (int j = 0; j < RunLists.size(); j++){
				Line = Line + String.valueOf(RunLists.get(j).get(i)) + "\t";
			}
			
			//print line!
			System.out.println(Line);
		}
		
		// BUMP COUNTS - looking for trends in the organism set
		// result = weak trend, towards host type (Paramecium Bursaria based)
		
//		LinkedHashMap<LinkedList<String>, Integer> OrgSetCounts = BumpCounts;
//		LinkedList<ListAndCounts> CountStats = new LinkedList<ListAndCounts>();
//		
//		for (LinkedList<String> L : BumpCounts.keySet()){
//			ListAndCounts LAC = new ListAndCounts();
//			LAC.List = L;
//			LAC.Count = BumpCounts.get(L);
//			CountStats.add(LAC);
//		}
//		
//		//sort!
//		Collections.sort(CountStats, new ListAndCountSorter());
		
		//System.out.println("Breakpoint!");
		
		//Export max Pairs
		//String MaxPairFile = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/JCE_Files/QS_LocalMaxPairs_Exact20.txt";
		//ExportLocalMaxPairs(MaxPairFile);
		
		
//		LinkedList<Integer> NewSet = new LinkedList<Integer>(AllClusters);
//		Collections.sort(NewSet);
//		
//		String str = "";
//		for (Integer x : NewSet){
//			str = str + String.valueOf(x) + "; ";
//		}
//		System.out.println(str);
		
		
		
		//Completion
		//System.out.println("All Processes Successfully Completed!");
	}
	
	public static class ListAndCounts {
		public LinkedList<String> List;
		public int Count;
	}
	
	public static class ListAndCountSorter implements Comparator<ListAndCounts>{

		@Override
		public int compare(ListAndCounts o1, ListAndCounts o2) {
			
			return (o1.Count - o2.Count);
		}
		
	}


}
