package FileFormatting;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.LinkedHashMap;
import java.util.LinkedList;

public class InvertedRepeats extends LoadData{

	/**
	 * @param args
	 */
	
	//Files
	public static String TIRFiles = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/JCE_Files/TIRFiles.txt";
	public static String LeftMotifs = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/JCE_Files/LeftTIRs.txt";
	public static String RightMotifs = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/JCE_Files/RightTIRs.txt";
	public static String GenomeFiles = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/JCE_Files/GenomeFiles.txt";
	
	//Data Fields
	public static LinkedList<String> TIRFileList;
	public static LinkedList<TIRInstance> TIRs;
	
	//constants
	public static int BaseBuffer = 1000;
	
	//classes
	public static class TIRInstance{
		public String Genome;
		public String Contig;
		public int LeftStart;
		public int LeftStop;
		public int LeftCenter;
		public int LeftSize;
		public int RightStart;
		public int RightStop;
		public int RightCenter;
		public int RightSize;

		}
	
	// ------ Import -------- //
	
	//Import TIRFiles
	public static void ImportTIRFiles(String TIRFiles){
		TIRFileList = new LinkedList<String>();
		try {
			BufferedReader br = new BufferedReader(new FileReader(TIRFiles));
			String Line = null;
			while ((Line = br.readLine()) != null){
				TIRFileList.add(Line);
			}
			br.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}

	}
	
	//Import a single TIR results file
	public static void ImportSingleTIRFile(String FileName){
		try {
			BufferedReader br = new BufferedReader(new FileReader(FileName));
			String Line = null;
			String OrgFile = "";
			String Contig = "";
			boolean InTheTable = false;
			while ((Line = br.readLine()) != null){
				
				//System.out.println(Line);
				
				//get sequence name
				if (Line.startsWith("Sequence")){
					Contig = Line.substring(10).trim();
					//System.out.println(Contig);
					OrgFile = OrgNames.get(Contig);
					
					if (OrgFile == null){
						if (Contig.contains("Reversed:")){
							String[] L = Contig.split(" ");
							Contig = L[0];
							OrgFile = OrgNames.get(Contig);
						}
					}
					
					if (Contig.equals("PBCV-1")){
						Contig = "PBCV1";
						OrgFile = OrgNames.get(Contig);
						Contig = "PBCV-1";
					}
					
					
					if (Contig.contains("MT325 TIGR assembly ID 72 314335 bp 23")){
						Contig = "MT325";
						OrgFile = OrgNames.get(Contig);
					}
					
					//error checking
					if (OrgFile == null){
						System.out.println(Contig);
					}

				}
				
				//while in the table, determine TIR instances
				if (InTheTable){
					
					//not an internal heading, not empty
					if (!Line.equals("") && !Line.startsWith("<TR><TD WIDTH=140>")){
						//System.out.println(Line);
						
						//New instance
						TIRInstance TIR = new TIRInstance();
						TIR.Genome = OrgFile;
						TIR.Contig = Contig;
						
						//extract left and right coordinates
						String[] LPound = Line.split("#");
						String[] LComma = LPound[1].split(",");
						
						String LeftSide = LComma[0];
						String RightSide = LComma[2];
						
						String[] LLeft = LeftSide.split("\\--");
						TIR.LeftStart = Integer.parseInt(LLeft[0]);
						TIR.LeftStop = Integer.parseInt(LLeft[1]);
						TIR.LeftCenter = (TIR.LeftStart + TIR.LeftStop)/2;
						TIR.LeftSize = TIR.LeftStop - TIR.LeftStart + 1;
						
						String[] LRight = RightSide.split("\\--");
						TIR.RightStart = Integer.parseInt(LRight[0]);
						TIR.RightStop = Integer.parseInt(LRight[1]);
						TIR.RightCenter = (TIR.RightStart + TIR.RightStop)/2;
						TIR.RightSize = TIR.RightStop - TIR.RightStart + 1;
						
//						//debugging - info
//						System.out.println(TIR.Genome + "\t"
//								+TIR.LeftStart + ":" + TIR.LeftStop + "\t"
//								+TIR.RightStart + ":" + TIR.RightStop);	

						//store in list
						TIRs.add(TIR);
						
					} else if (Line.equals("")){
						InTheTable = false;
					}

				}
				
				
				//into the table
				if (Line.startsWith("<TR><TD WIDTH=140>")){
					InTheTable = true;
				}
				
				
			}
			br.close();
		} catch (Exception ex){
			ex.printStackTrace();
			System.exit(1);
		}
	}
	
	// ------- Export --------- //
	
	//Export left
	public static void ExportLeft(String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			for (TIRInstance TIR : TIRs){
				String str = TIR.Genome + "\t" 
						+ TIR.Contig + "\t"
						+ String.valueOf(TIR.LeftStart) + "\t"
						+ String.valueOf(TIR.LeftStop) + "\tTIR_Left\n"
						;
				bw.write(str);
				bw.flush();
				//System.out.println(str);
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export right
	public static void ExportRight(String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			for (TIRInstance TIR : TIRs){
				String str = TIR.Genome + "\t" 
						+ TIR.Contig + "\t"
						+ String.valueOf(TIR.RightStart) + "\t"
						+ String.valueOf(TIR.RightStop) + "\tTIR_Right\n"
						;
				bw.write(str);
				bw.flush();
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export all regions where a TIR is within 4K to 5K as context set
	public static void ExportTIRContextSet(String FileName, String DataFileName, int StartRange, int StopRange, int MinSize){
		
		try {
			//open file
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			BufferedWriter bw2 = new BufferedWriter(new FileWriter(DataFileName));
			int Counter = 0;
			
			//iterate through every TIR, check for space limitations
			for (TIRInstance TIR : TIRs){
				
				//distance between TIR centers is in range + both TIRs large enough
				if (Math.abs(TIR.LeftCenter-TIR.RightCenter) > StartRange &&
						Math.abs(TIR.LeftCenter-TIR.RightCenter) < StopRange &&
						Math.min(TIR.LeftSize,TIR.RightSize) >= MinSize
						){
					
					//increment counter
					Counter++;
					
					//find all genes between centers, + buffer (1000 nt)
					int MinimalStart = Math.min(TIR.LeftCenter, TIR.RightCenter) - BaseBuffer;
					int MaximalStop = Math.max(TIR.LeftCenter, TIR.RightCenter) + BaseBuffer;
					
					//retrieve features
					//System.out.println(Chloroviruses.get(TIR.Genome));
					LinkedList<GenomicFeature> Features = Chloroviruses.get(TIR.Genome);
				
					//find features around these TIRs
					for (GenomicFeature GF : Features){
						
						//all genes within the appropriate range.
						if (GF.StartPos >= MinimalStart && GF.StopPos <= MaximalStop){
							
							String str = GF.OrgName + "\t" + GF.Contig + "\t"
									+String.valueOf(GF.StartPos) + "\t" + String.valueOf(GF.StopPos)
									+"\t1\n";
							bw.write(str);
							bw.flush();
							
						}
					}
					
					//write the whole range
					String str = TIR.Genome + "\t" + MinimalStart + "\t" + MaximalStop + "\n";
					bw2.write(str);
					
				}
				
			}
			
			//message
			System.out.println(Counter + " TIRs match specifications.");
			
			//close file streams
			bw.close();
			bw2.close();
			
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	// ----- Processing ------- //
	
	
	//main method
	public static void main(String[] args) {
		
		//Import org names
		OrgNames = new LinkedHashMap<String, String>();
		TIRs = new LinkedList<TIRInstance>();
		
		LoadAllData();
		//ImportOrgNames();
		
		//Import data
		ImportTIRFiles(TIRFiles);
		for (String s : TIRFileList){
			ImportSingleTIRFile(s);
		}
		
		//debugging
//		int counter = 0;
//		for (TIRInstance TIR : TIRs){
//			counter++;
//			System.out.println(counter + "\t" + TIR.Contig + "\t" + TIR.LeftCenter + "\t" + TIR.RightCenter);
//		}
		
//		//Export Left + Right motif files
//		ExportLeft(LeftMotifs);
//		ExportRight(RightMotifs);
		
		//Export Context Set
		String TIRCS = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/Miscellaneous/TIRCS.txt";
		String TIRCS_Data = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/Miscellaneous/TIRCS_Data.txt";
		ExportTIRContextSet(TIRCS,TIRCS_Data,4000,12000,250);
		//output message
		System.out.println("All Processes Successfully Completed!");
		
	}

}
