package FileFormatting;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map.Entry;

import FileFormatting.LoadData.GenomicFeature;

public class MultipleMemberGeneGangs extends LoadData{

	/**
	 * @param args
	 */
	public static String BaseDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs";
	public static String AmalgDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/AmalgamatedGangs";
	public static String ClusterGangDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/ClusterGangs";
	public static String GGListDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/GeneLists";
	public static String TableDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/PBCV1Tables";
	public static String GeneGangContextSetDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/ContextSets";
	public static String StatisticsDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/Statistics";
	public static String ProtSeqDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/GeneGangProteinSequences";
	public static String SingleClusterDir = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/ClusterProteinSequences";

	//Cast nets threshold + unruliness threshold
	public static int CastNetsThreshold = 15;
	public static double UScale;
	public static LinkedHashMap<ParameterSet, AmalgamatedTableRow> Results 
		= new LinkedHashMap<ParameterSet, AmalgamatedTableRow>();
	public static LinkedList<ParameterSet> Parameters;
	
	//count-related
	public static int CoreGenesInGeneGangs = 0;
	public static int NonLargeCoreGenesInGeneGangs = 0;
	public static int NonLargeNonCoreGenesInGeneGangs = 0;
	public static LinkedHashMap<Double,GeneGangStats> RulinessGGCountMapping;
	//cluster gang - a summary of search results

	// ======= Classes ======== //

	public static class ClusterGang{
		public int SeedCluster;
		public LinkedList<Integer> ClustersFeatured;
		public LinkedList<LinkedList<Integer>> AllClusters;
		public LinkedList<Integer> GeneGangClusters;
		
		//This is the key! organism, feature mappings
		public LinkedHashMap<String, LinkedList<GenomicFeature>> ClusterOrgMapping;
		
		//10-17-2013: 		 org name		all segments that correlate!
		public LinkedHashMap<String, LinkedList<LinkedList<GenomicFeature>>> ClusterOrgMappingUnMerged;
		
		//10-17-2013: large cluster instances
		public LinkedHashMap<String, LinkedHashMap<Integer, Integer>> LargeClusterInstancesMerged;
		
		//10-17-2013: all features
		public LinkedList<GenomicFeature> AllFeatures;
	}
	
	//object for output table -> comprehensive statistics
	public static class AmalgamatedTableRow{
		public int Range;
		public double Unruliness;
		public double Ruliness;
		public LinkedList<LinkedList<Integer>> AmalgamatedGangs;
		public LinkedHashMap<Integer, Integer> Counts;		//GG Size, Number of occurrences in whole set
		public AmalgamatedGeneGang AGG;
	}
	
	//object for amalgamated grouping
	public static class AmalgamatedGeneGang{
		public LinkedList<LinkedList<Integer>> ClusterIDs;
		public LinkedHashMap<LinkedList<Integer>, LinkedList<GenomicFeature>> GenomicFeaturesByClusterMembers;
		
		//retrieve by gang number
		public LinkedHashMap<Integer, LinkedList<GenomicFeature>> GenomicFeaturesByGangNumber;
		public LinkedHashMap<Integer, LinkedHashMap<String, LinkedList<GenomicFeature>>> GeneGangsByOrg;
		public LinkedHashMap<Integer, LinkedList<Integer>> ClustersFeatured;
		
		public Double Right;
		public Double Below;
		public Double AverageDiss;
		public Double PreAmalgamatedUnruliness;
		public Double PostAmalgamatedUnruliness;
		public LinkedList<Double> PostAmalgamatedUnrulinessRange;
		
		public int Range;
		public int AmalgamationConvergenceIterations;
		
		//counts			 GangSize,Num of Features
		public LinkedHashMap<Integer,Integer> GangSizeCounts;
		
		//for use with parameter determination
		public LinkedList<LinkedList<GenomicFeature>> PBCV1_LargeGG;
		
		//for use with final, post-amalgamation unruliness
		public LinkedHashMap<GenomicFeature, Double> MaximumUnrulinessMap;
		
	}
	
	//Objects for statistics
	public static class GeneGangStats{
		public double Ruliness;
		public int Range;
		public LinkedHashMap<Integer,Integer> NonLargeGangSizeCounts;
		public LinkedHashMap<Integer,Integer> GangSizeCounts;
		public LinkedHashMap<Integer,Integer> NonLargeCoreGangSizeCounts;
		public LinkedHashMap<Integer,Integer> CoreGangSizeCounts;
	}
	
	//object for parameter set
	public static class ParameterSet {
		public int Range;
		public double Ruliness;
		public double Unruliness;
	}
	
	//sort a list of elements
	public static class ListSorter implements Comparator<LinkedList<Integer>>{

		@Override
		public int compare(LinkedList<Integer> L1, LinkedList<Integer> L2) {
			return L1.get(0) - L2.get(0);
		}
		
	}
	
	//sort a list of integers, based on size
	public static class SizeSorter implements Comparator<LinkedList<Integer>>{

		@Override
		public int compare(LinkedList<Integer> L1, LinkedList<Integer> L2) {
			if (L1.size() == L2.size()){
				return L1.get(0) - L2.get(0);
			} else {
				return (L2.size() - L1.size());
			}
		}
		
	}
	
	//sort a list of genomic feature
	public static class GFListSorter implements Comparator<GenomicFeature>{

		@Override
		public int compare(GenomicFeature F1, GenomicFeature F2) {
			return F1.CenterPos - F2.CenterPos;
		}
		
	}
	
	//sort a list of lists
	public static class LLSorter implements Comparator<LinkedList>{

		@Override
		public int compare(LinkedList arg0, LinkedList arg1) {
			// TODO Auto-generated method stub
			return -1*(arg0.size() - arg1.size());
		}
		
	}
	
	//sort parameter set
	public static class ParameterSetSorter implements Comparator<ParameterSet> {

		@Override
		public int compare(ParameterSet PS1, ParameterSet PS2) {
			
			// TODO Auto-generated method stub
			return 0;
		}
		
	}
	
	// ====== Whole Pipeline ====== //
	
	//Multiple Gene Gangs - return gene gang breakdown, given the rangearound and unruliness.
	public static AmalgamatedTableRow MultipleGeneGangs(int RangeAround, double Ruliness, double Unruliness, String FileName){

		//for debugging + display.
		System.out.println("-------------------");
		System.out.println("Range: " + RangeAround + " Ruliness: " + Ruliness + " Unruliness: " + Unruliness);
		
		//Initialize output + variables
		//Table row
		AmalgamatedTableRow ATR = new AmalgamatedTableRow();
		ATR.Range = RangeAround;
		ATR.Unruliness = Unruliness;
		ATR.Ruliness = Ruliness;
		CoreGenesInGeneGangs = 0;
		
		//File names
		//String AmalgFile = AmalgDir + "/" + FileName;
		String ClusterGangFile = ClusterGangDir + "/" + FileName;
		//String GeneGangsAsIDFile = GGListDir + "/" + FileName;
		//String GeneGangContextSetFile = GeneGangContextSetDir + "/" + FileName;
		//String SummaryTableFile = TableDir + "/" + FileName;
		
		// ====== Determine Cluster Gangs ======== //
		
		//Cluster Gangs using the 'paralog compression' approach, and avoiding large clusters.
		LinkedList<ClusterGang> ClusterGangs = RetrieveAllGroupingsAndCompressParalogs(RangeAround);
		
		// ====== Compress to gene gangs ========//
		
		//determine cluster gangs - the gene gang, as it exists from the perspective of a single cluster
		LinkedList<ClusterGang> ClusterGangsMod = ComputeIndividualClusterGangs(ClusterGangs,Ruliness,RangeAround);
		
		//System.out.println("Computed all individual gangs.  Entering amalgamation step.");
		
		//Greedy amalgamation - join clusters, with the exception of large cluster (multiple paralog problem).
		AmalgamatedGeneGang GeneGangs = GreedyAmalgamation2(ClusterGangsMod,RangeAround,Ruliness);

		//Correct for errors associated with erroneous exclusion of multiple paralogs
		//GeneGangs = CastNets(GeneGangs, 3500);	//I just set this value because it works
		GeneGangs = CastNets(GeneGangs, RangeAround/2);
		
		//write values
		GeneGangs.Range = RangeAround;
		GeneGangs.PreAmalgamatedUnruliness = Unruliness;
		
		ATR.AGG = GeneGangs;
		ATR.AmalgamatedGangs = GeneGangs.ClusterIDs;
		
		//Compute counts, based on these gangs.
		ATR.Counts = GeneGangSizeCounts(GeneGangs.ClusterIDs);

		//debugging - hold off on export
		//(1) Gene gangs from a particular perspective (cluster-based gangs)
		ExportClusterGeneGangGroups(ClusterGangsMod, ClusterGangFile);
		
		//(2) Amalgamated gene gangs (from greedy algorithm, excluding large clusters)
		//ExportAmalgamatedGeneGangs(GeneGangs.ClusterIDs, AmalgFile);
		
		//(3) Gene gangs as lists of IDs instead of cluster numbers
		//ExportGeneGangGeneIDs(GeneGangs, GeneGangsAsIDFile);
		
		//(4) Genomic Groupings (Context Set) of gene gangs
		//ExportGeneGangContextSets(GeneGangs, GeneGangContextSetFile);
		
		//(5) Summary output table
		//ExportPBCV1SummaryTable(GeneGangs, SummaryTableFile);
		
		//final message
		System.out.println("-------------------");
		
		//output
		return ATR;

	}
	
	// ====== Parameter statistics type measures ====== //
	
	//counts for statistics
	//Count gene gangs of various size
	public static LinkedHashMap<Integer, Integer> GeneGangSizeCounts(LinkedList<LinkedList<Integer>> GeneGangs){
		
		//Initialize output
		LinkedHashMap<Integer, Integer> Counts = new LinkedHashMap<Integer, Integer>();
		
		//count values
		for (LinkedList<Integer> L : GeneGangs){
			int x = L.size();
			if (x > 10){
				x = 11;
			}
			if (Counts.get(x) != null){
				int Val = Counts.get(x);
				Val++;
				Counts.put(x,Val);
			} else {
				Counts.put(x, 1);
			}
		}
		
		//System.out.println("Breakpoint!");
		
		//return
		return Counts;
	}

	// ====== Amalgamations ====== // 
	
	//Greedy amalgamation - merge non-Large clusters
	public static AmalgamatedGeneGang GreedyAmalgamation(LinkedList<ClusterGang> ClusterGangsMod, int RangeAround, double Unruliness){
		
		/*
		 * Greedy amalgamations: if the intersection is not zero, merge!
		 */
		
		//Initialize outputs
		AmalgamatedGeneGang AGG = new AmalgamatedGeneGang();
		LinkedList<LinkedList<Integer>> GeneGangs = new LinkedList<LinkedList<Integer>>();
		
		boolean FirstGang = true;
		boolean NewGeneGang = false;
		boolean FoundMatch = false;
		
		for (ClusterGang CG : ClusterGangsMod){
			if (FirstGang){
				GeneGangs.add(CG.GeneGangClusters);
				FirstGang = false;
			} else{
				
				NewGeneGang = true;
				FoundMatch = false;
				//check every existing gene gang, merge if any overlaps
				for (LinkedList<Integer> x : GeneGangs){
					for (Integer y : CG.GeneGangClusters){
						if (x.contains(y) && !LargeClusters.contains(y)){
							x.addAll(CG.GeneGangClusters);
							NewGeneGang = false;
							FoundMatch = true;
							break;
						}
					}
					if (FoundMatch){
						break;
					}
				}
				
				//if no overlaps, 
				if (NewGeneGang){
					GeneGangs.add(CG.GeneGangClusters);
				}

			}
		}
		
		//No duplicates
		LinkedList<LinkedList<Integer>> GeneGangsNoDups = new LinkedList<LinkedList<Integer>>();
		
		for (LinkedList<Integer> x : GeneGangs){
			LinkedList<Integer> Gang = new LinkedList<Integer>();
			for (Integer y : x){
				if (!Gang.contains(y)){
					Gang.add(y);
				}
			}
			GeneGangsNoDups.add(Gang);
		}
		
		//sort
		for (LinkedList<Integer> x : GeneGangsNoDups){
			Collections.sort(x);
		}
		Collections.sort(GeneGangsNoDups, new ListSorter());
		Collections.sort(GeneGangsNoDups, new SizeSorter());
		
		//store value
		AGG.ClusterIDs = GeneGangsNoDups;
		
		//Initialize a map
		LinkedHashMap<LinkedList<Integer>, LinkedList<GenomicFeature>> IDElementMapping = new
				LinkedHashMap<LinkedList<Integer>, LinkedList<GenomicFeature>>();
		
		//check every gang
		for (LinkedList<Integer> x : GeneGangsNoDups){
		
			//re-initialize features list + hash map
			LinkedList<GenomicFeature> GGFeatures = new LinkedList<GenomicFeature>();
			LinkedHashMap<String, LinkedList<GenomicFeature>> FeaturesHash =
						new LinkedHashMap<String, LinkedList<GenomicFeature>>();
			
			//every gene within the gang
			for (Integer y : x){

				//search all available profiles
				for (ClusterGang CG : ClusterGangsMod){
					
					//find the match
					if (CG.SeedCluster == y){
						
						//retrieve features, add to new hash
						for (String s : CG.ClusterOrgMapping.keySet()){
							LinkedList<GenomicFeature> LL = CG.ClusterOrgMapping.get(s);
							
							if (FeaturesHash.get(s) != null){
								LinkedList<GenomicFeature> newList = FeaturesHash.get(s);
								for (GenomicFeature GF : LL){
									if (!newList.contains(GF)){
										newList.add(GF);
									}
								}
								FeaturesHash.put(s,newList);
							} else {
								FeaturesHash.put(s,LL);
							}
						}
												
						break;
					}
				}

			}
			
			//amalgamate the hash map
			for (String s : FeaturesHash.keySet()){
				
//				//debugging
//				if (s.equals("Fr5L") && x.contains(120) && Unruliness == 0.19 && RangeAround == 15000){
//					System.out.println("Breakpoint");
//				}
				
				LinkedList<GenomicFeature> Features = FeaturesHash.get(s);
				LinkedList<GenomicFeature> UpdatedFeatures = CollinearFilter(Features,RangeAround);
				
				//Test - add filtering step for removing stray genes that are far from the others
				UpdatedFeatures = FarAwayLargeFilter(UpdatedFeatures, 3000);
				
				GGFeatures.addAll(UpdatedFeatures);
			}
			
			//write to hash map
			IDElementMapping.put(x, GGFeatures);
		}
	
		//store value
		AGG.GenomicFeaturesByClusterMembers = IDElementMapping;
		
		return AGG;
		
	}
		
	//Greedy amalgamation - gene-centric, no gene may be in multiple gangs
	public static AmalgamatedGeneGang GreedyAmalgamation2(LinkedList<ClusterGang> ClusterGangsMod, int RangeAround, double Ruliness){
		
		/*
		 * Algorithmic design: a featured-based (gene-centric) approach
		 * key: a feature may never be in more than one gene gang - if that is ever the case, then merge.
		 * keep merging until there are 0 intersections.  Even a single gene intersection entails merging
		 * the whole gang.
		 */
		
		/*
		 * ==========================
		 * PART 1: Iterative amalgamation
		 * ==========================
		 */
		
		//Initialize output + parameters
		AmalgamatedGeneGang AGG = new AmalgamatedGeneGang();
		boolean GenesAllInUnique = false;
		
		//reset counters
		NonLargeCoreGenesInGeneGangs = 0;
		NonLargeNonCoreGenesInGeneGangs = 0;
		
		//Mapping of each gene gang to all genes, in all 41 chloroviruses, in the gang
		LinkedHashMap<Integer, LinkedList<GenomicFeature>> GeneGangMapping = new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
				
		//initialize gene gang mapping
		int Counter = 0;
		for (ClusterGang CG : ClusterGangsMod){
			Counter++;
			GeneGangMapping.put(Counter, CG.AllFeatures);			
		}
		
		//iteration counter - for convergence.
		int IterationCounter = 0;
		
		//output message
		System.out.println("Iteration 0: " + GeneGangMapping.size() + " gangs.");
		
		//Keep merging until no need to merge any more
		while (!GenesAllInUnique){

			//Variables to be re-initialized with every iteration.
			LinkedHashMap<Integer, LinkedList<Integer>> MergingSchedule = new LinkedHashMap<Integer, LinkedList<Integer>>();
			LinkedList<Integer> KeepUnMerged = new LinkedList<Integer>();
			
			//Increment iteration counter
			IterationCounter++;
			
			//check every gang against every other gang
			for (Integer x : GeneGangMapping.keySet()){
				
				//scan this gang against every other gang
				for (Integer y : GeneGangMapping.keySet()){
					if (x != y){
						
						//retrieve lists
						LinkedList<GenomicFeature> List1 = GeneGangMapping.get(x);
						LinkedList<GenomicFeature> List2 = GeneGangMapping.get(y);
						
						//look for intersections
						boolean FoundCommonFeature = false;
						for (GenomicFeature GF1 : List1){
							for (GenomicFeature GF2 : List2){
								if (GF1.GeneID.equals(GF2.GeneID)){
								//if (GF1.equals(GF2)){
									FoundCommonFeature = true;
									break;
								}
							}
							if (FoundCommonFeature){
								break;
							}
						}
						
						//when intersections found, schedule for merging
						if (FoundCommonFeature){
							if (MergingSchedule.get(x) != null){
								LinkedList<Integer> MergeWith = MergingSchedule.get(x);
								if (!MergeWith.contains(y)){
									MergeWith.add(y);
								}
								MergingSchedule.put(x, MergeWith);
							} else {
								LinkedList<Integer> MergeWith = new LinkedList<Integer>();
								MergeWith.add(y);
								MergingSchedule.put(x, MergeWith);
							}
						}

					}
				}
				
			}

			//check merging schedule
			GenesAllInUnique = true;
			for (Integer q : GeneGangMapping.keySet()){
				if (MergingSchedule.get(q) != null){
					
					//convergence not reached.
					GenesAllInUnique = false;
					
				} else {
					//note the original sets (singletons)
					KeepUnMerged.add(q);
				}
			}
			
			//create new set
			if (!GenesAllInUnique){
			
				//Initialize set for merging
				LinkedList<LinkedList<Integer>> GeneGangs4Merge = new LinkedList<LinkedList<Integer>>();
				
				//create the merge set
				for (Integer x : MergingSchedule.keySet()){
					
					//blob of gangs to add
					LinkedList<Integer> MergeWithX = MergingSchedule.get(x);
					MergeWithX.add(x);
					
					//add this new blob by itself, or merge with an existing??
					boolean AddNewList = true;
					
					//check existing list, modify if necessary
					for (LinkedList<Integer> Lists : GeneGangs4Merge){
						for (Integer p : Lists){
							if (MergeWithX.contains(p)){
								AddNewList = false;
								break;
							}
						}

						if (!AddNewList){
							for (Integer p : MergeWithX){
								if (!Lists.contains(p)){
									Lists.add(p);
								}
							}
							break;
						}
					}
					
					//if still no intersection, add as new unit
					if (AddNewList){
						GeneGangs4Merge.add(MergeWithX);
					}
					
				}

				//Initialize new gene gang structure
				LinkedHashMap<Integer, LinkedList<GenomicFeature>> NewGeneGangMapping = new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
				
				//fill out - add new gangs
				int NewGangCounter = 0;
				for (LinkedList<Integer> Gangs2Merge : GeneGangs4Merge){
					
					//Initialize condensed list
					HashSet<GenomicFeature> Merged = new HashSet<GenomicFeature>();
					
					for (Integer x : Gangs2Merge){
						//retrieve features + add to hash table
						LinkedList<GenomicFeature> Features = GeneGangMapping.get(x);
						Merged.addAll(Features);
					}
					
					//convert to linked list
					LinkedList<GenomicFeature> MergedGFs = new LinkedList<GenomicFeature>(Merged);
					
					//store
					NewGangCounter++;
					NewGeneGangMapping.put(NewGangCounter,MergedGFs);
				}
				
				//add old gangs
				for (Integer x : KeepUnMerged){
					LinkedList<GenomicFeature> Features = GeneGangMapping.get(x);
					NewGangCounter++;
					NewGeneGangMapping.put(NewGangCounter, Features);
				}
				
				//update
				GeneGangMapping = NewGeneGangMapping;
				
				//message - in regards to older version
				System.out.println("Iteration " + IterationCounter + ": " + GeneGangMapping.size() + " gangs.");
				
			} else {
				IterationCounter--;
				System.out.println("Amalgamation process converged after " + IterationCounter + " iteration(s).");
				AGG.AmalgamationConvergenceIterations = IterationCounter;
			}
			
		}
		
		/*
		 * ==========================
		 * PART 2: Final collection
		 * ==========================
		 */
		
		int PreFilterCounter = 0;
		
		//Initialize
		LinkedHashMap<Integer, LinkedList<GenomicFeature>> FilteredHashMap = 
				new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
		
		//build org-specific.
		LinkedHashMap<Integer, LinkedHashMap<String, LinkedList<GenomicFeature>>> GeneGangsByOrg 
			= new LinkedHashMap<Integer, LinkedHashMap<String, LinkedList<GenomicFeature>>>();
		
		//build clusters featured in gangs.
		LinkedHashMap<Integer, LinkedList<Integer>> ClustersInGangs
			= new LinkedHashMap<Integer, LinkedList<Integer>>();
		
		//colinear filtering - only appropriate for species-specific!
		for (Integer x : GeneGangMapping.keySet()){
			
			//Initialize
			LinkedList<GenomicFeature> FilteredFeatures = new LinkedList<GenomicFeature>();
			LinkedList<Integer> ClusterList = new LinkedList<Integer>();
			
			//Retrieve list of genomic features
			LinkedList<GenomicFeature> Features = GeneGangMapping.get(x);
			PreFilterCounter = PreFilterCounter + Features.size();
		
			//segment by species
			LinkedHashMap<String, LinkedList<GenomicFeature>> BySpecies
				= new LinkedHashMap<String, LinkedList<GenomicFeature>>();
			
			for (GenomicFeature GF : Features){
				if (BySpecies.get(GF.OrgName) != null){
					LinkedList<GenomicFeature> CurrentFs = BySpecies.get(GF.OrgName);
					CurrentFs.add(GF);
					BySpecies.put(GF.OrgName, CurrentFs);
				} else {
					LinkedList<GenomicFeature> CurrentFs = new LinkedList<GenomicFeature>();
					CurrentFs.add(GF);
					BySpecies.put(GF.OrgName, CurrentFs);
				}
			}
			
			//filtered hash map
			LinkedHashMap<String, LinkedList<GenomicFeature>> GeneGangXFilteredHash =
					new LinkedHashMap<String, LinkedList<GenomicFeature>>();
			
			//filter + amalgamate
			for (String s : BySpecies.keySet()){
				
				//Colinear filter
				LinkedList<GenomicFeature> OrgSpecificFeatures = BySpecies.get(s);
				
				//Mod 10-21-2013: let's just forget about collinear filtering (at least for now)
				//Mod 10-22-2013: Colinear filtering is back
				OrgSpecificFeatures = CollinearFilter(OrgSpecificFeatures, RangeAround);
				
				//add to hash map
				GeneGangXFilteredHash.put(s, OrgSpecificFeatures);
				
				//add to master list
				FilteredFeatures.addAll(OrgSpecificFeatures);
				
				//add clusters
				for (GenomicFeature GF : OrgSpecificFeatures){
					if (!ClusterList.contains(GF.ClusterNum)){
						ClusterList.add(GF.ClusterNum);
					}
				}
			}
					
			//update hashmap with all genes
			FilteredHashMap.put(x,FilteredFeatures);

			//update species-divided hashmap
			GeneGangsByOrg.put(x, GeneGangXFilteredHash);
			
			//update cluster counts hash map
			Collections.sort(ClusterList);
			ClustersInGangs.put(x,ClusterList);
		}
		
		//store new filtered
		AGG.GeneGangsByOrg = GeneGangsByOrg;
		AGG.ClustersFeatured = ClustersInGangs;
		
		//switch to filtered
		GeneGangMapping = FilteredHashMap;
		
		//store mapping
		AGG.GenomicFeaturesByGangNumber = GeneGangMapping;

		//Cluster IDs + Features hash
		LinkedList<LinkedList<Integer>> ClusterIDs = new LinkedList<LinkedList<Integer>>();
		LinkedHashMap<LinkedList<Integer>,LinkedList<GenomicFeature>> GenomicFeatures =
				new LinkedHashMap<LinkedList<Integer>,LinkedList<GenomicFeature>>();
		
		/*
		 * ==========================
		 * PART 3: Counting
		 * ==========================
		 */
		
		//counts
		for (Integer x : GeneGangMapping.keySet()){
			
			//retrieve features
			LinkedList<GenomicFeature> Features = GeneGangMapping.get(x);
			
			//initialize
			LinkedList<Integer> Clusters = new LinkedList<Integer>();

			//retrieve main clusters from list
			for (GenomicFeature GF : Features){
				//store in list
				if (GF.ClusterNum != 0 && !Clusters.contains(GF.ClusterNum)){
					Clusters.add(GF.ClusterNum);
				}
				
				//retrieve data
				ClusterData CD = ClusterDirectoryHash.get(GF.ClusterNum);
				
				//count based on types
				if (!CD.isLargeCluster){
					if (CD.isCore){
						NonLargeCoreGenesInGeneGangs++;
						CoreGenesInGeneGangs++;
					} else {
						NonLargeNonCoreGenesInGeneGangs++;
					}
				} else {
					CoreGenesInGeneGangs++;
				}

			}
					
			//sort list
			Collections.sort(Clusters);
			
			//store in list
			if (!ClusterIDs.contains(Clusters)){
				ClusterIDs.add(Clusters);
			}
			
			//map these clusters with features
			GenomicFeatures.put(Clusters, Features);
		}
				
		//sort list appropriately
		Collections.sort(ClusterIDs, new ListSorter());
		Collections.sort(ClusterIDs, new SizeSorter());
		
		//update mappings 
		AGG.GenomicFeaturesByClusterMembers = GenomicFeatures;
		AGG.ClusterIDs = ClusterIDs;
		
		/*
		 * Statistics
		 */

		GeneGangStats GGS = new GeneGangStats();
		GGS.Ruliness = Ruliness;
		GGS.Range = RangeAround;
		
//						//All genes in total 	  num in gene gangs (including single member gangs)
//		int Leftovers =  NonLargeCoreGeneCount - NonLargeCoreGenesInGeneGangs;
//		
//		
//		//output message
//		System.out.println("Import count: " + NonLargeCoreGeneCount 
//				+ ", Nonlarge in gang count: " + NonLargeCoreGenesInGeneGangs 
//				+ ", Leftovers: " + Leftovers);
		
		//new computation 10-30-2013
		int Leftovers = (NonLargeNonCoreGeneCount+NonLargeCoreGeneCount)
								- (NonLargeNonCoreGenesInGeneGangs+NonLargeCoreGenesInGeneGangs);
		
		//output message
		System.out.println("Import count: " + (NonLargeNonCoreGeneCount+NonLargeCoreGeneCount)
				+ ", Nonlarge in gang count: " + (NonLargeNonCoreGenesInGeneGangs+NonLargeCoreGenesInGeneGangs) 
				+ ", Leftovers: " + Leftovers);
		
		//Store counts
		LinkedHashMap<Integer,Integer> GangSizeCounts = new LinkedHashMap<Integer,Integer>();
		LinkedHashMap<Integer,Integer> NonLargeGangSizeCounts = new LinkedHashMap<Integer,Integer>();
		
		GangSizeCounts.put(0, Leftovers);
		NonLargeGangSizeCounts.put(0, Leftovers);
		
		//determine counts from sets of cluster IDs
		for (LinkedList<Integer> L : ClusterIDs){
			
			//retrieve features associated with this cluster
			LinkedList<GenomicFeature> Features = GenomicFeatures.get(L);
			
			//determine the number of non-large core genes in the list.
			int CoreCounter = 0;
			for (GenomicFeature GF : Features){
				if (!LargeClusters.contains(GF.ClusterNum)){
					CoreCounter++;
				}
			}
			
			//add to hash map
			if (GangSizeCounts.get(L.size()) == null){
				GangSizeCounts.put(L.size(), CoreCounter);
			} else {
				int Counts = GangSizeCounts.get(L.size());
				Counts = Counts + CoreCounter;
				GangSizeCounts.put(L.size(), Counts);
			}
			
			//determine alternative counts
			int NonLargeSize = 0;
			for (Integer x : L){
				if (!LargeClusters.contains(x)){
					NonLargeSize++;
				}
			}

			//add to hash map
			if (NonLargeGangSizeCounts.get(NonLargeSize) == null){
				NonLargeGangSizeCounts.put(NonLargeSize, CoreCounter);
			} else {
				int Counts = NonLargeGangSizeCounts.get(NonLargeSize);
				Counts = Counts + CoreCounter;
				NonLargeGangSizeCounts.put(NonLargeSize, Counts);
			}
		}
		
		GGS.GangSizeCounts = GangSizeCounts;
		GGS.NonLargeGangSizeCounts = NonLargeGangSizeCounts;
		
		//store value
		RulinessGGCountMapping.put(Ruliness,GGS);
		
		
//		//Debugging: find identity of leftovers
//		LinkedList<GenomicFeature> Features = Chloroviruses.get("PBCV1");
//		
//		for (GenomicFeature GF : Features){
//			boolean Found = false;
//			for (LinkedList<GenomicFeature> LL: AGG.GenomicFeatureByGang.values()){
//				if (LL.contains(GF)){
//					Found = true;
//					break;
//				}
//			}
//			
//			//display
//			if (!Found){
//				if (GF.ClusterNum != 0 && !LargeClusters.contains(GF.ClusterNum)){
//					System.out.println(GF.GeneID);
//				}
//			}
//		}		
		
		//return output
		return AGG;
	}
	
	//Cast Nets (modify amalgamated set to include possible misses)
	public static AmalgamatedGeneGang CastNets(AmalgamatedGeneGang AGG, Integer NetSize){
		
		/*
		 * The purpose of this step is to catch any genes that may have been excluded
		 * during the Greedy Amalgamation, largely because they are part of large clusters.
		 * Procedure:
		 * (1) Identify all clusters associated with a gene gang.
		 * (2) Look for additional cluster members before / after the edges
		 *     of a post-processed gene gang + schedule for merging.
		 * (3) modify the existing gang with the merged set.
		 */
		
		System.out.println("Casting Nets.");
		
		//Initialize a schedule
		LinkedHashMap<Integer, LinkedList<GenomicFeature>> ScheduledForMerging
			= new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
		
		//for every gene gang
		for (Integer x : AGG.GeneGangsByOrg.keySet()){
			
			//Initialize list of schedules
			LinkedList<GenomicFeature> Schedule = new LinkedList<GenomicFeature>();
			
			//retrieve hash of features
			LinkedHashMap<String, LinkedList<GenomicFeature>> FeatureHash = AGG.GeneGangsByOrg.get(x);
			
			//retrieve list of clusters
			LinkedList<Integer> Clusters = AGG.ClustersFeatured.get(x);
			
			//iterate through species
			for (String s : FeatureHash.keySet()){
				
				//retrieve list of features
				LinkedList<GenomicFeature> Features = FeatureHash.get(s);
				
				//retrieve start/stop of whole block
				int StartNet = Features.get(0).StartPos - NetSize;
				int StopNet = Features.getLast().StopPos + NetSize;
				
				//check genes
				LinkedList<GenomicFeature> OrgGenes = Chloroviruses.get(s);
				
				//identify all features that should be scheduled for merging
				for (GenomicFeature GF : OrgGenes){
					if (GF.CenterPos > StartNet
							&& GF.CenterPos < StopNet
							&& Clusters.contains(GF.ClusterNum)
							&& !Features.contains(GF)){
						Schedule.add(GF);
					}
				}
			}
			
			//store results
			ScheduledForMerging.put(x,Schedule);
			
		}
		
//		//debugging!
//		for (Integer x : ScheduledForMerging.keySet()){
//			if (ScheduledForMerging.get(x).size() > 0){
//				System.out.println(x + ": " 
//						+ AGG.ClustersFeatured.get(x) + " " 
//						+ ScheduledForMerging.get(x).size() + " to add.");
//			}
//		}
		
		//modify the master list, according to cast nets threshold.
		for (Integer x : AGG.GenomicFeaturesByGangNumber.keySet()){
			if (ScheduledForMerging.get(x).size() > CastNetsThreshold){
				LinkedList<GenomicFeature> FeatureList = AGG.GenomicFeaturesByGangNumber.get(x);
				FeatureList.addAll(ScheduledForMerging.get(x));
				AGG.GenomicFeaturesByGangNumber.put(x, FeatureList);
			}
		}
		
		//return
		return AGG;
	}
	
	//Determine best set, based on breaking point.
	public static AmalgamatedGeneGang PreAmalgamatedBreakingPoint(LinkedList<AmalgamatedGeneGang> AGGList){
		
		/*
		 * The definition of 'unruliness' changes when looking pre- and post- amalgamation.
		 * This approach aims to determine the maximum pre-amalgamation unruliness
		 * where all gangs stay together.
		 * From this point, these gangs can be re-evaluated for a post-amalgamation unruliness
		 * (which should resemble the intuitive definition for unruliness).
		 */
		
		System.out.println("Adjusting post-amalgamated gangs by unruliness.");
		
		AmalgamatedGeneGang AGGtoReturn= null;		
		for (AmalgamatedGeneGang AGG : AGGList){
			
			//break out of loop, if necessary.
			if (AGG.AmalgamationConvergenceIterations > 1){
				break;
			}
			AGGtoReturn = AGG;
		}
		
		//return
		return AGGtoReturn;
	}
	
	//re-classify gene gangs based on post-amalgamated unruliness
	public static AmalgamatedGeneGang DeterminePostAmalgamatedUnruliness(AmalgamatedGeneGang AGG){
		
		 //Initialize
		 AGG.MaximumUnrulinessMap = new LinkedHashMap<GenomicFeature, Double>();
		
		 //range of values
		 LinkedList<Double> UnrulinessRange = new LinkedList<Double>();
		 
		 //check every gang
		 for (Integer x : AGG.GeneGangsByOrg.keySet()){
			 
			 //retrieve info
			 LinkedHashMap<String, LinkedList<GenomicFeature>> FeaturesByOrg = AGG.GeneGangsByOrg.get(x);
			 LinkedList<Integer> ClustersFeatured = AGG.ClustersFeatured.get(x);
			 
			 //debugging - ensure that all gangs involve 41 bugs
			 if (ClustersFeatured.size() >= 3){
				 System.out.println("Gang: " + x + ", orgs: " + FeaturesByOrg.size());
			 }
			 
			 //for each cluster, count occurrences
			 for (Integer y : ClustersFeatured){
				 
				 //reset counter for each cluster
				 int OrgCount = 0;
				 
				 //list of collected features for each cluster (across organisms)
				 LinkedList<GenomicFeature> FeaturesToMark = new LinkedList<GenomicFeature>();
				 
				 //iterate through each organism
				 for (String s : FeaturesByOrg.keySet()){
					 
					 //did this organism contain one or more instances of the cluster?
					 boolean OneOrMore = false;
					 
					 //retrieve features in this organism
					 LinkedList<GenomicFeature> Features = FeaturesByOrg.get(s);

					 //search until the feature is found in this organism, if it ever is.
					 for (GenomicFeature GF : Features){
						 if (GF.ClusterNum == y){
							 FeaturesToMark.add(GF);
							 OneOrMore = true;
						 }
					 }
					 
					 //update counter, if appropriate
					 if (OneOrMore){
						 OrgCount++;
					 }
					 
				 }
				 
				 //total org count -> map to an unruliness value.
				 double FU = (double) OrgCount / (double) Chloroviruses.size();
				 FU = (double)Math.round(FU * 1000) / 1000;
				 
				 //add features to hash map
				 for (GenomicFeature GF : FeaturesToMark){
					 AGG.MaximumUnrulinessMap.put(GF, FU);
					 if (!UnrulinessRange.contains(FU) && ClustersFeatured.size() >= 3){
						 UnrulinessRange.add(FU);
					 }
				 }
			 }
		 }
		 
		 //debugging: sort + Display featured unruliness values
		 Collections.sort(UnrulinessRange);
		 for (Double d : UnrulinessRange){
			 System.out.println(d);
		 }
		 
		 //note the range
		 AGG.PostAmalgamatedUnrulinessRange = UnrulinessRange;
		 
		 //return gene gang with info
		 return AGG;
	}
	
	//segment a gene gang based on classified unruliness
	public static AmalgamatedGeneGang RecomputeBasedOnUnruliness(AmalgamatedGeneGang AGG, Double d){
					
		//set stage for re-importation
		AmalgamatedGeneGang AGG_Filtered = AGG;
		LinkedHashMap<Integer, LinkedList<GenomicFeature>> NewMapping = new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
		LinkedHashMap<Integer, LinkedList<Integer>> NewClusterIDs = new LinkedHashMap<Integer, LinkedList<Integer>>();

		//build hash map of features values
		for (Integer x : AGG.GenomicFeaturesByGangNumber.keySet()){
			
			//lists of features
			LinkedList<GenomicFeature> OriginalList = AGG.GenomicFeaturesByGangNumber.get(x);
			LinkedList<GenomicFeature> PrunedList = new LinkedList<GenomicFeature>();
			
			//new clusters featured, given the threshold
			LinkedList<Integer> PrunedClustersFeatured = new LinkedList<Integer>();
			
			//add approrpiate features to new list
			for (GenomicFeature GF : OriginalList){
				if (AGG.MaximumUnrulinessMap.get(GF) >= d){
					PrunedList.add(GF);
					
					//add the new cluster ID to the list
					if (!PrunedClustersFeatured.contains(GF.ClusterNum)){
						PrunedClustersFeatured.add(GF.ClusterNum);
					}
				}
			}
			
			//sort list
			Collections.sort(PrunedClustersFeatured);
			
			//store data in new hash maps
			NewMapping.put(x, PrunedList);
			NewClusterIDs.put(x, PrunedClustersFeatured);
			
		}
		
		//store this agglomeration into the map
		//AGG_Filtered.GenomicFeaturesByGangNumber = null;
		AGG_Filtered.GenomicFeaturesByGangNumber = NewMapping;
		AGG_Filtered.ClustersFeatured = NewClusterIDs;
		AGG_Filtered.PostAmalgamatedUnruliness = 1.000 - d;
		return AGG_Filtered;
		
	}
	
	//determine gene gang stats
	public static GeneGangStats PostAmalgamationStats(AmalgamatedGeneGang AGG){
		
		//Initialize
		GeneGangStats GGS = new GeneGangStats();
		
		//set parameter values
		GGS.Ruliness = 1.000-AGG.PostAmalgamatedUnruliness;
		GGS.Range = AGG.Range;
		
		//create hash maps
		LinkedHashMap<Integer, Integer> CountWithLarge = new LinkedHashMap<Integer, Integer>();
		LinkedHashMap<Integer, Integer> CountNoLarge = new LinkedHashMap<Integer, Integer>();
		LinkedHashMap<Integer,Integer> CoreWithLarge = new LinkedHashMap<Integer,Integer>();
		LinkedHashMap<Integer,Integer> CoreNoLarge = new LinkedHashMap<Integer,Integer>();

		//count based on present/absent clusters
		for (Integer x : AGG.ClustersFeatured.keySet()){
			
			//determine number of non-large genes of this size
			int NonLargeGenesInGang = 0;
			int NonLargeCoreGenesInGang = 0;
			LinkedList<GenomicFeature> Features = AGG.GenomicFeaturesByGangNumber.get(x);
			for (GenomicFeature GF : Features){
				if (!LargeClusters.contains(GF.ClusterNum)){
					NonLargeGenesInGang++;
				}
				if (ClusterDirectory.get(GF.ClusterNum).isCore){
					NonLargeCoreGenesInGang++;
				}
			}
			
			//generate lists of clusters (size of list -> number of gene members)
			LinkedList<Integer> ListwLarge = AGG.ClustersFeatured.get(x);
			LinkedList<Integer> ListnoLarge = new LinkedList<Integer>();
			for (Integer y : ListwLarge){
				if (!LargeClusters.contains(y)){
					ListnoLarge.add(y);
				}
			}
			
			/*
			 * Counts
			 */
			
			//all
			//include large genes in count
			if (CountWithLarge.get(ListwLarge.size()) != null){
				int Val = CountWithLarge.get(ListwLarge.size());
				Val = Val + NonLargeGenesInGang; //all genes in this gang
				CountWithLarge.put(ListwLarge.size(),Val);
			} else {
				CountWithLarge.put(ListwLarge.size(), NonLargeGenesInGang);
			}
			
			//exclude large genes from count
			if (CountNoLarge.get(ListnoLarge.size()) != null){
				int Val = CountNoLarge.get(ListnoLarge.size());
				Val = Val + NonLargeGenesInGang; //all genes in this gang
				CountNoLarge.put(ListnoLarge.size(),Val);
			} else {
				CountNoLarge.put(ListnoLarge.size(), NonLargeCoreGenesInGang);
			}
			
			//core
			//include large genes in count - core
			if (CoreWithLarge.get(ListwLarge.size()) != null){
				int Val = CoreWithLarge.get(ListwLarge.size());
				Val = Val + NonLargeCoreGenesInGang; //core genes in this gang
				CoreWithLarge.put(ListwLarge.size(),Val);
			} else {
				CoreWithLarge.put(ListwLarge.size(), NonLargeCoreGenesInGang);
			}
			
			//exclude large genes from count-core
			if (CoreNoLarge.get(ListnoLarge.size()) != null){
				int Val = CoreNoLarge.get(ListnoLarge.size());
				Val = Val + NonLargeCoreGenesInGang; //core genes in this gang
				CoreNoLarge.put(ListnoLarge.size(),Val);
			} else {
				CoreNoLarge.put(ListnoLarge.size(), NonLargeCoreGenesInGang);
			}
			
			
		}
		
		//store counts
		GGS.GangSizeCounts = CountWithLarge;
		GGS.NonLargeGangSizeCounts = CountNoLarge;
		GGS.CoreGangSizeCounts = CoreWithLarge;
		GGS.NonLargeCoreGangSizeCounts = CoreNoLarge;
		
		//return
		return GGS;
		
	}
	
	// ======= Components ======== //
	
	//Compute individual gene gangs, starting from a seed cluster
	public static LinkedList<ClusterGang> ComputeIndividualClusterGangs(LinkedList<ClusterGang> LL, double Ruliness, int RangeAround){
		
		//output?
		LinkedList<ClusterGang> NewGangs = new LinkedList<ClusterGang>();
		
		//compute for each
		for (ClusterGang CG : LL){
			
			LinkedList<Integer> GG = new LinkedList<Integer>();
			
			//Add clusters that are appropriate to the gang, according to 'unruliness' specification.
			for (Integer x : CG.ClustersFeatured){
				
				//reset counter
				int OrgCount = 0;
				
				//count instances
				for (LinkedList<Integer> List : CG.AllClusters){
					if (List.contains(x)){
						OrgCount++;
					}
				}
				
				//determine if sufficient to retain
				double Frac = (double) OrgCount / (double) CG.AllClusters.size();
				
				//comparison
				if (Frac >= Ruliness){
					GG.add(x);
				}
			}
			
			//remove duplicates + sort
			LinkedList<Integer> Gang = new LinkedList<Integer>();
			for (Integer y : GG){
				if (!Gang.contains(y)){
					Gang.add(y);
				}
			}
			Collections.sort(Gang);
			
			//update with gg hits
			CG.GeneGangClusters = Gang;
			
			//All features
			LinkedList<GenomicFeature> AllFeatures = new LinkedList<GenomicFeature>();
			
			//update for each organism.
			for (String s : CG.ClusterOrgMapping.keySet()){
				
				//Initialize filtered list
				LinkedList<GenomicFeature> FilteredList = new LinkedList<GenomicFeature>();
				
				//Retrieve unfiltered list
				LinkedList<GenomicFeature> UnfilteredList = CG.ClusterOrgMapping.get(s);
				
				//(1) Build list based on presence of appropriate cluster
				for (Integer x : Gang){
					for (GenomicFeature GF : UnfilteredList){
						if (GF.ClusterNum == x){
							FilteredList.add(GF);
						}
					}
				}
				
				//colinear filter
				LinkedList<GenomicFeature> CoFilteredList = CollinearFilter(FilteredList,RangeAround);
				
				//update
				CG.ClusterOrgMapping.put(s, CoFilteredList);
				
				//add these features
				AllFeatures.addAll(CoFilteredList);

			}
			
			//Note all features
			CG.AllFeatures = AllFeatures;
			
			//debugging - don't add empty gangs
			if (CG.AllFeatures.size() > 0){
				NewGangs.add(CG);
			}
			
			//add to new list?
			//NewGangs.add(CG);
		}
		
//		//debugging
//		for (ClusterGang CG : NewGangs){
//			System.out.println(CG.SeedCluster + "\t" + CG.AllFeatures.size());
//		}
		
		return NewGangs;
	}

	//determine all cluster gangs + compress paralogs
	public static LinkedList<ClusterGang> RetrieveAllGroupingsAndCompressParalogs(int RangeAround){
	
		//Initialize hash map
		LinkedList<ClusterGang> Output= new LinkedList<ClusterGang>();
		
		//Retrieve gangs for each species, ignoring cluster 1
		//old: iterate from i = 2 to 155
		//
		//new: iterate over whole set, but ignore large clusters
		for (int i = 1; i <= LargestCluster; i++){
			
			if (!LargeClusters.contains(i)){
				
				ClusterGang CG = new ClusterGang();
				CG.SeedCluster = i;
				
				//Store 
				LinkedList<LinkedList<Integer>> ClusterIDs = new LinkedList<LinkedList<Integer>>();
				LinkedHashMap<String, LinkedList<GenomicFeature>> ClusterIDOrgMapping = new LinkedHashMap<String, LinkedList<GenomicFeature>>();
				LinkedHashMap<String, LinkedList<LinkedList<GenomicFeature>>> ClusterOrgMappingUnMerged = new LinkedHashMap<String, LinkedList<LinkedList<GenomicFeature>>>();
				LinkedList<Integer> InAllClusters = new LinkedList<Integer>();
				LinkedHashMap<String, LinkedHashMap<Integer, Integer>> LargeClusterInstancesMerged = new LinkedHashMap<String, LinkedHashMap<Integer, Integer>>();
				//all candidate genomic groupings for each cluster ID
				for (String s : Chloroviruses.keySet()){
					
					//Retrieve clusters for this feature
					LinkedList<LinkedList<GenomicFeature>> OutputList = RetrieveGenomicGrouping(s,RangeAround,i);
					
					//recover all cluster IDs featured
					LinkedList<Integer> ClustersFeatured = new LinkedList<Integer>();
					LinkedList<GenomicFeature> ClustersAndIDsFeatured = new LinkedList<GenomicFeature>();
					
					//keep genomic features separate
					ClusterOrgMappingUnMerged.put(s, OutputList);
					
					//keep all hits
					for (LinkedList<GenomicFeature> Segment : OutputList){
						for (GenomicFeature GF : Segment){
							
							//only retain core genes
							if (GF.ClusterNum != 0){
								//add clusters as appropriate
								
								//add instances.
								if (!ClustersAndIDsFeatured.contains(GF)){
									ClustersAndIDsFeatured.add(GF);
								}

								//only need to add cluster numbers once.
								if (!ClustersFeatured.contains(GF.ClusterNum)){
									ClustersFeatured.add(GF.ClusterNum);
								}
								if (!InAllClusters.contains(GF.ClusterNum)){
									InAllClusters.add(GF.ClusterNum);
								}
								
								//counts of large clusters
								if (LargeClusters.contains(GF.ClusterNum)){
									if (LargeClusterInstancesMerged.get(s) != null){
										LinkedHashMap<Integer,Integer> CountsForOrg = LargeClusterInstancesMerged.get(s);
										if (CountsForOrg.get(GF.ClusterNum) != null){
											int Count = CountsForOrg.get(GF.ClusterNum);
											Count++;
											CountsForOrg.put(GF.ClusterNum, Count);
										} else {
											CountsForOrg.put(GF.ClusterNum,1);
										}
										LargeClusterInstancesMerged.put(s, CountsForOrg);
									} else {
										LinkedHashMap<Integer,Integer> CountsForOrg = new LinkedHashMap<Integer,Integer>();
										CountsForOrg.put(GF.ClusterNum, 1);
										LargeClusterInstancesMerged.put(s, CountsForOrg);
									}
								}
								
							}
						}
					}
					
					//note counts
					
					
					//add these clusters to the list
					ClusterIDs.add(ClustersFeatured);
					ClusterIDOrgMapping.put(s, ClustersAndIDsFeatured);
					
				}
				
				//complete cluster gang
				CG.AllClusters = ClusterIDs;
				CG.ClustersFeatured = InAllClusters;
				CG.ClusterOrgMapping = ClusterIDOrgMapping;	//organism, LinkedList<GenomicFeature>
				CG.ClusterOrgMappingUnMerged = ClusterOrgMappingUnMerged; 	//unmerged - cases with multiple segments
				
				//add to hash map
				Output.add(CG);
			}

		}
		
		return Output;
	}
	
	//retrieve an appropriate genomic grouping (list of genomic features)
	public static LinkedList<LinkedList<GenomicFeature>> RetrieveGenomicGrouping(String OrgName, int RangeAround, int ClusterNum){
		
		//Initialize output - a list of lists of all regions (the features themselves, in order)
		LinkedList<LinkedList<GenomicFeature>> ClusterIDs = new LinkedList<LinkedList<GenomicFeature>>();
		
		//Retrieve features
		LinkedList<GenomicFeature> Features = Chloroviruses.get(OrgName);
		
		//list of matching indices.
		LinkedList<GenomicFeature> MatchingIndexes = new LinkedList<GenomicFeature>();
		LinkedHashMap<GenomicFeature, Integer> Feature2Position = new LinkedHashMap<GenomicFeature, Integer>();
		
		//find the key feature
		for (int i = 0; i < Features.size(); i++){
			if (Features.get(i).ClusterNum == ClusterNum){
				MatchingIndexes.add(Features.get(i));
				Feature2Position.put(Features.get(i), i);
			}
		}
		
		//iterate through all discovered hits
		for (GenomicFeature SeedFeature : MatchingIndexes){
			
			LinkedList<GenomicFeature> IDwRange = new LinkedList<GenomicFeature>();
			int Seed = Feature2Position.get(SeedFeature);	//seed position
			IDwRange.add(SeedFeature);
			int HalfRange = (int) Math.round(0.5* (double) RangeAround);
			int i = Seed; 		//start at seed
			
			//Before Seed
			boolean KeepAddingFeatures = true;
			while (KeepAddingFeatures) {
				
				//boundary
				if (i == Features.size()-1 || i == 0){
					KeepAddingFeatures = false;
				} else {
					
					//check for distance
					if (Features.get(Seed).CenterPos-Features.get(i).CenterPos <= HalfRange){
						if (Features.get(i).ClusterNum != 0 && Features.get(i) != SeedFeature){
							IDwRange.add(0,Features.get(i));
						}
						i--;
					} else {
						KeepAddingFeatures = false;
					}
					
				}

			}
			
				//After Seed
				i = Seed;	//re-set to seed
				KeepAddingFeatures = true;
				while (KeepAddingFeatures) {
					
					//boundary
					if (i == Features.size()-1 || i == 0){
						KeepAddingFeatures = false;
					} else {
						
						//check for distance
						if (Features.get(i).CenterPos-Features.get(Seed).CenterPos <= HalfRange){
							if (Features.get(i).ClusterNum != 0 && Features.get(i) != SeedFeature){
								IDwRange.add(Features.get(i));
							}
							i++;
						} else {
							KeepAddingFeatures = false;
						}
						
					}

				}
			
			ClusterIDs.add(IDwRange);

		}
		
		//output 
		return ClusterIDs;
		
	}
	
	// ======= Filters ======= //

	//Winnow a list of potential candidates based on colinearity (large gaps between gene blocks)
	public static LinkedList<GenomicFeature> CollinearFilter(LinkedList<GenomicFeature> InitialList, int RangeAround){
	
		/*
		 * Algorithm:
		 * (1) Sort elements in the list, in order of appearance on the genome
		 * (2) If the center of an element is within (RangeAround) from the next element,
		 * 	   keep the elements together.  Otherwise, create a new list, and put this element with the new list.
		 * (3) Return the grouping with the largest number of members. 
		 */
		
		if (InitialList.size() > 0){
			
			//sort elements in list
			Collections.sort(InitialList, new GFListSorter());

//			if (InitialList.get(0).StartPos == 215470 &&
//					InitialList.get(0).StopPos == 215883 &&
//					InitialList.get(0).Contig.equals("MA-1E")){
//				
//				System.out.println("Breakpoint!");
//				
//			}
			
			//initialize output list
			LinkedList<LinkedList<GenomicFeature>> SegmentedList = new LinkedList<LinkedList<GenomicFeature>>();
			
			//initialize: first set
			LinkedList<GenomicFeature> L = new LinkedList<GenomicFeature>();
			L.add(InitialList.get(0));
			
			for (int i = 1; i < InitialList.size(); i++){
				
				//element belongs with this list
				if (InitialList.get(i).CenterPos - InitialList.get(i-1).CenterPos < RangeAround){
					L.add(InitialList.get(i));
				} else {
					SegmentedList.add(L);
					L = new LinkedList<GenomicFeature>();
					L.add(InitialList.get(i));
				}
			}
			
			//add last list
			SegmentedList.add(L);
			
			//store lists according to counts
			LinkedHashMap<Integer, LinkedList<GenomicFeature>> Counts = new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
			int MaxCount = 0;
			
			//check all lists
			for (LinkedList<GenomicFeature> SegL : SegmentedList){
				
				//adjust max count
				if (SegL.size() > MaxCount){
					MaxCount = SegL.size();
				}
				
				//store - either augment an existing list, or create a new list
				if (Counts.get(SegL.size()) == null){
					Counts.put(SegL.size(), SegL);
				} else {
					
					//old - add all, new, just replace with later instance
					//Counts.get(SegL.size()).addAll(SegL);
					Counts.put(SegL.size(), SegL);
				}
			}
			
			//return max count
			return Counts.get(MaxCount);
		} else {
			return InitialList;
		}

	}

	//Winnow a list of potential candidates based on far-away members in the "large" cluster set
	public static LinkedList<GenomicFeature> FarAwayLargeFilter(LinkedList<GenomicFeature> InitialList, int FarAwayRange){
				
		/*
		 * Algorithm:
		 * (1) Sort elements in the list, in order of appearance on the genome
		 * (2) If the center of an element is within (RangeAround) from the next element,
		 * 	   keep the elements together.  Otherwise, create a new list, and put this element with the new list.
		 * (3) Return the grouping with the largest number of members. 
		 */
		
		//sort elements in list
		Collections.sort(InitialList, new GFListSorter());

		//initialize output list
		LinkedList<LinkedList<GenomicFeature>> SegmentedList = new LinkedList<LinkedList<GenomicFeature>>();
		
		//initialize: first set
		LinkedList<GenomicFeature> L = new LinkedList<GenomicFeature>();
		L.add(InitialList.get(0));
		
		for (int i = 1; i < InitialList.size(); i++){
				
			//Consider Large Clusters
			//possibility for splitting - a large cluster is involved.
			if (LargeClusters.contains(InitialList.get(i).ClusterNum) ||
					LargeClusters.contains(InitialList.get(i-1).ClusterNum)){
				
				//split only if far away
				if (InitialList.get(i).StartPos - InitialList.get(i-1).StopPos > FarAwayRange){
					SegmentedList.add(L);
					L = new LinkedList<GenomicFeature>();
					L.add(InitialList.get(i));
				} else {
					L.add(InitialList.get(i));
				}
				
			} else {
				L.add(InitialList.get(i));
			}
			
//			
//			//Do Not Consider Large Clusters
//			//element belongs with this list
//			if (InitialList.get(i).StartPos - InitialList.get(i-1).StopPos < FarAwayRange){
//				L.add(InitialList.get(i));
//			} else {
//				SegmentedList.add(L);
//				L = new LinkedList<GenomicFeature>();
//				L.add(InitialList.get(i));
//			}
		}
		
		//add last list
		SegmentedList.add(L);
		
		//store lists according to counts
		LinkedHashMap<Integer, LinkedList<GenomicFeature>> Counts = new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
		int MaxCount = 0;
		
		//check all lists
		for (LinkedList<GenomicFeature> SegL : SegmentedList){
			
			//adjust max count
			if (SegL.size() > MaxCount){
				MaxCount = SegL.size();
			}
			
			//store - either augment an existing list, or create a new list
			if (Counts.get(SegL.size()) == null){
				Counts.put(SegL.size(), SegL);
			} else {
				Counts.get(SegL.size()).addAll(SegL);
			}
		}
		
		//return max count
		return Counts.get(MaxCount);
		
	}
	
	// ====== Print Output ======= //
	
	//display gene gangs (non-amalgamted, cluster-centric)
	public static void PrintClusterGangGeneGangs(int RangeAround, double Unruliness, LinkedList<ClusterGang> ClusterGangs){
		
		System.out.println("Conserved Cluster Gangs");
		System.out.println("RangeAround = " + String.valueOf(RangeAround) + " Unruliness = " + String.valueOf(Unruliness));
		
		String str = "";
		//print results
		for (ClusterGang CG : ClusterGangs){
			str = CG.SeedCluster + ": ";
			for (Integer x : CG.GeneGangClusters){
				str = str + x + ", ";
			}
			System.out.println(str);
		}
		
	}
	
	//display gene gangs (amalgamated)
	public static void PrintGeneGangs(LinkedList<LinkedList<Integer>> GeneGangs){
		String str = "";
		for (LinkedList<Integer> x : GeneGangs){
			str = "";
			for (Integer y : x){
				str = str + String.valueOf(y) + "; ";
			}
			System.out.println(str);
		}
	}
	
	//print ATR matrix
	public static void PrintATRMatrix(AmalgamatedTableRow[][] ATRMatrix){
		
		//System.out.println(ATRMatrix.length + "\t" + ATRMatrix[0].length);
		
		for (int i = 0; i < ATRMatrix.length; i++){
			String MainRow = "";
			String SubRow = "";
			for (int j = 0; j < ATRMatrix[i].length; j++){
				MainRow = MainRow + "(" + ATRMatrix[i][j].Range + "," + ATRMatrix[i][j].Ruliness + ")" + "\t";

				//add below diss
				if (i < ATRMatrix.length - 1){
					SubRow = SubRow + ATRMatrix[i][j].AGG.Below + "\t";
				}
				
				//add right-ward diss
				if (j < ATRMatrix[i].length - 1){
					MainRow = MainRow + ATRMatrix[i][j].AGG.Right + "\t";
					SubRow = SubRow + "\t\t";
				}
				
			}
			System.out.println(MainRow);
			System.out.println(SubRow);
		}
	}
	
	//Jaccard comparison metric - original
	public static Double CompareAmalgamations(AmalgamatedGeneGang AGG1, AmalgamatedGeneGang AGG2){
		
		//initialize output + counters
		double Dissimilarity = 0.0;
		double NumIntersecting = 0;
		double SizeAGG1;
		double SizeAGG2;
		double SizeUnion = 0;

		//Hash Sets
		
		//just second set
		HashSet<LinkedList<Integer>> O2Hash = new HashSet<LinkedList<Integer>>(AGG2.ClusterIDs);
		
		//intersection
		HashSet<LinkedList<Integer>> IntersectionHash = new HashSet<LinkedList<Integer>>(AGG1.ClusterIDs);
		IntersectionHash.retainAll(O2Hash);
		
		//union
		HashSet<LinkedList<Integer>> UnionHash = new HashSet<LinkedList<Integer>>(AGG1.ClusterIDs);
		UnionHash.addAll(O2Hash);
		
		//sizes
		SizeAGG1 = AGG1.ClusterIDs.size();
		SizeAGG2 = AGG2.ClusterIDs.size();

		//Find all intersecting types, and find the number that intersect.
		for (LinkedList<Integer> O : IntersectionHash){
			NumIntersecting = NumIntersecting + Math.min(Collections.frequency(AGG1.ClusterIDs, O), Collections.frequency(AGG2.ClusterIDs, O));
		}

		//compute union
		SizeUnion = SizeAGG1 + SizeAGG2 - NumIntersecting;
		
		if (SizeUnion != 0) {
			Dissimilarity =  1-(NumIntersecting/SizeUnion);
		} else { //divide by zero case
			Dissimilarity = 0;
		}
		
		//dissimilarity
		return Dissimilarity;
		
	}

	//Generalized Dice/Jaccard (returns similarity)
	public static double GeneralizedDiceOrJaccard(LinkedList<GenomicFeature> O1, LinkedList<GenomicFeature> O2, boolean TreatDuplicatesAsUnique, String Type){
		
		//Initialize values
		double Dissimilarity = 0;
		double Similarity = 0;
		double NumIntersecting = 0;
		double SizeO1;
		double SizeO2;
		double SizeUnion = 0;
		
		//Hash Sets
		HashSet<GenomicFeature> O1Hash = new HashSet<GenomicFeature>(O1);
		HashSet<GenomicFeature> O2Hash = new HashSet<GenomicFeature>(O2);
		HashSet<GenomicFeature> IntersectionHash = new HashSet<GenomicFeature>(O1);
		HashSet<GenomicFeature> UnionHash = new HashSet<GenomicFeature>(O1);
		IntersectionHash.retainAll(O2Hash);
		UnionHash.addAll(O2Hash);
		
		if (TreatDuplicatesAsUnique){
			SizeO1 = O1.size();
			SizeO2 = O2.size();

			//Find all intersecting types, and find the number that intersect.
			for (Object O : IntersectionHash){
				NumIntersecting = NumIntersecting + Math.min(Collections.frequency(O1, O), Collections.frequency(O2, O));
			}
			
			//compute union
			SizeUnion = SizeO1 + SizeO2 - NumIntersecting;
			
		} else {

			SizeO1 = O1Hash.size();
			SizeO2 = O2Hash.size();
			NumIntersecting = IntersectionHash.size();
			SizeUnion = UnionHash.size();
		}

		//compute dissimilarity from computed size / union / intersection values
		if (Type.equals("Dice")){
			if (!((SizeO1 == 0) && (SizeO2 == 0))){
				Dissimilarity = 1-(2*NumIntersecting)/(SizeO1+SizeO2);
				Similarity = 1-Dissimilarity;
			} else { //divide by zero case
				Dissimilarity = 0;
				Similarity = 0;
			}
		} else {	//Jaccard
			if (SizeUnion != 0) {
				Dissimilarity =  1-(NumIntersecting/SizeUnion);
				Similarity = NumIntersecting/SizeUnion;
			} else { //divide by zero case
				Dissimilarity = 0;
				Similarity = 0;
			}
			
		}

		return Similarity;
	}
		
	//Jaccard comparison metric - PBCV1 gene lists (returns similarity)
	public static Double CompareLargePBCV1Amalgamations(AmalgamatedGeneGang AGG1, AmalgamatedGeneGang AGG2){
		
		//recover lists
		LinkedList<LinkedList<GenomicFeature>> LL1 = AGG1.PBCV1_LargeGG;
		LinkedList<LinkedList<GenomicFeature>> LL2 = AGG2.PBCV1_LargeGG;
		
		//Total similarity
		double TotalSimilarity = 0;
		
		//for every list, find closest matching list
		for (LinkedList<GenomicFeature> L1 : LL1 ){
			double MaxClosest = 0;
			for (LinkedList<GenomicFeature> L2 : LL2){
//				double Comparison = JaccardFeatureList(L1,L2);
				double Comparison = GeneralizedDiceOrJaccard(L1,L2,true,"Dice");
				if (Comparison > MaxClosest){
					MaxClosest = Comparison;
				}
			}
			TotalSimilarity = TotalSimilarity + MaxClosest;
		}
		
		//determine average best similarity
		return TotalSimilarity / (double) LL1.size();
		
	}
	
	//Display mappings
	public static void PrintGangMappings(double Ruliness){
		
//		System.out.println("Ruliness = " + Ruliness);
//		
//		LinkedHashMap<Integer,Integer> Mappings = RulinessGGCountMapping.get(Ruliness);
//		
//		for (Integer x : Mappings.keySet()){
//			int Val = Mappings.get(x);
//			
//			double Frac = (double) Val / (double) CoreGeneCount ;
//			Frac = 100 * Frac;
//			double RoundedFrac =(double)Math.round(Frac * 10) / 10;
//			
//			String str = "Size " + x + ": " + Mappings.get(x) 
//					+ "/" + CoreGeneCount + " or " + String.valueOf(RoundedFrac) 
//					+ "% of core genes in gene gangs.";
//			
//			System.out.println(str);
//		}
	}
	
	// ======= Parsing =========//
	
	//re-format geneID
	public static String JCEReformatGeneID(GenomicFeature GF){
		
		//Initial gene ID
		String s = GF.GeneID;
		
		//make replacements
		s = s.replace("_Ref","");
		s = s.replace('|','_');
		
		//return modified gene ID
		return s;
	}
	
	// ======= Export ========= //
	
	//Export list of cluster IDs featured in amalgamated gene gangs. - not an AGG
	public static void ExportAmalgamatedGeneGangs(LinkedList<LinkedList<Integer>> GeneGangs, String FileName){
		
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			String str = "";
			for (LinkedList<Integer> x : GeneGangs){
				str = "";
				for (Integer y : x){
					str = str + String.valueOf(y) + "; ";
				}
				str = str + "\n";
				bw.write(str);
				bw.flush();
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export a list of cluster IDs associated with amalgamted gene gangs.
	public static void ExportAmalgamatedGeneGangs2(AmalgamatedGeneGang AGG, String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			String str = "";
			for (Integer z : AGG.ClustersFeatured.keySet()){
				LinkedList<Integer> x = AGG.ClustersFeatured.get(z);
				str = "";
				for (Integer y : x){
					str = str + String.valueOf(y) + "; ";
				}
				str = str + "\n";
				bw.write(str);
				bw.flush();
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
		
	//Export cluster-centric gene gangs (as cluster numbers) - not an AGG
	public static void ExportClusterGeneGangGroups(LinkedList<ClusterGang> ClusterGangs, String FileName){
		
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			String str = "";
			//print results
			for (ClusterGang CG : ClusterGangs){
				str = CG.SeedCluster + ":\t";
				//str = "";
				//Edit Sep 3
				for (Integer x : CG.GeneGangClusters){
				//for (Integer x : CG.ClustersFeatured){
					str = str + x + "; ";
				}
				str = str + "\n";
				bw.write(str);
				bw.flush();
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export Gene Gangs as list of gene IDs instead of cluster IDs. - AGG.GenomicFeaturesByGangNumber
	public static void ExportGeneGangGeneIDs(AmalgamatedGeneGang AGG, String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			String str = "";
			
			//iterate through gangs
			for (Integer x : AGG.GenomicFeaturesByGangNumber.keySet()){
				
				//re-set string
				str = "";
				
				//iterate through features
				LinkedList<GenomicFeature> y = AGG.GenomicFeaturesByGangNumber.get(x);
				for (GenomicFeature GF : y){
					str = str + JCEReformatGeneID(GF) + "; ";
				}

				//add new line and export
				str = str + "\n";
				bw.write(str);
				bw.flush();
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export a PBCV-1 gene gang summary table. - AGG.GenomicFeaturesByGangNumber
	public static AmalgamatedGeneGang ExportPBCV1SummaryTable(AmalgamatedGeneGang GeneGangs, String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));			
			
			//Export Header
			String Header = "Gang\tPBCV-1_ID\tAnnotations\tCluster_ID\tExpression/Virion-Association\n";
			String Underline = "-------------------------------------------------------------" +
					"----------------------------------------------------------\n";
			bw.write(Header);
			bw.write(Underline);
			bw.flush();

			//Initialize list of lists
			LinkedList<LinkedList<GenomicFeature>> PBCV1GeneCollection = new LinkedList<LinkedList<GenomicFeature>>();
			
			//large groups only
			LinkedList<LinkedList<GenomicFeature>> PBCV1_LargeGG = new LinkedList<LinkedList<GenomicFeature>>();
			
			//mod 10/20/2013
			for (Integer x : GeneGangs.GenomicFeaturesByGangNumber.keySet()){
				
				//retrieve features
				LinkedList<GenomicFeature> Features = GeneGangs.GenomicFeaturesByGangNumber.get(x);
				LinkedList<GenomicFeature> PBCV1Genes = new LinkedList<GenomicFeature>();
				
				//retrieve PBCV-1 genes
				for (GenomicFeature GF : Features){
				if (GF.Contig.equals("PBCV-1")){
						PBCV1Genes.add(GF);
					}
				}
				
				//sort
				Collections.sort(PBCV1Genes, new GFListSorter());	
				
				//add to list
				PBCV1GeneCollection.add(PBCV1Genes);
				
				// if large enough, add to the large list
				if (PBCV1Genes.size() > 2){
					PBCV1_LargeGG.add(PBCV1Genes);
				}
				
			}
			
			//sort lists
			Collections.sort(PBCV1GeneCollection, new LLSorter());
			Collections.sort(PBCV1_LargeGG, new LLSorter());
			
			//prepare to return this value
			GeneGangs.PBCV1_LargeGG = PBCV1_LargeGG;
			
			//Variables for file writing
			int GangCounter = 0;
			String Entry = "";
			
			//write sorted list to file
			for (LinkedList<GenomicFeature> PBCV1Genes : PBCV1GeneCollection){
				
				//increment counter
				GangCounter++;
				
				for (GenomicFeature GF : PBCV1Genes){
					Entry = String.valueOf(GangCounter) + "\t" + JCEReformatGeneID(GF) + "\t" 
						+ ClusterAnnotations.get(GF.ClusterNum) + "\t"
						+ String.valueOf(GF.ClusterNum) + "\t\t\t\n";
					bw.write(Entry);
					bw.flush();
				}
				
				//new line between entries
				bw.write("\n");
				bw.flush();	
				
			}			
			
			//close file stream
			bw.close();
			
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		
		//return statement
		return GeneGangs;
	}
	
	//Export Context Set of Gene Gangs - AGG.GenomicFeaturesByGangNumber
	public static void ExportGeneGangContextSets(AmalgamatedGeneGang AGG, String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			String str = "";
			int Counter = 0;
			
			//iterate through gangs
			for (Integer x : AGG.GenomicFeaturesByGangNumber.keySet()){
				
				//increment counter
				Counter++;
				
				//iterate through features
				LinkedList<GenomicFeature> y = AGG.GenomicFeaturesByGangNumber.get(x);
				
				for (GenomicFeature GF : y){
					str = OrgNames.get(GF.OrgName) + "\t" + GF.Contig 
							+ "\t" + GF.StartPos 
							+ "\t" + GF.StopPos 
							+ "\t" + String.valueOf(Counter) + "\n";
					bw.write(str);
					bw.flush();
				}
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export Statistics - simple counts
	public static void ExportGeneGangStatsv2(String OutputFileName, LinkedHashMap<Double, GeneGangStats> GGMapping){
		
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(OutputFileName));
			
			//Iterate
			for (Double d : GGMapping.keySet()){
				
				//retrieve data
				GeneGangStats GGS = GGMapping.get(d);
				
				//Write intro
				String Header = "Ruliness = " + String.valueOf(d) + "\n";
				bw.write(Header);
				bw.flush();
				
				/*
				 * GROUP 1 : Gang Size is computed including large clusters
				 */
				
				//header
				String header1 = "Gang size includes large clusters:\n";
				bw.write(header1);
				bw.flush();
			
				//3 or more large clusters, not in a gene gang
				int ThreeOrMoreAll = 0;
				int ThreeOrMoreCore = 0;
				int NotInGeneGangAll = 0;
				int NotInGeneGangCore = 0;
				
				LinkedHashMap<Integer,Integer> Mappings = GGS.GangSizeCounts;
				for (int x = 100; x >= 0; x--){
					if (Mappings.get(x) != null){
						
						//Retrieve / Compute values
						int AllVal = Mappings.get(x);
						int CoreVal = GGS.CoreGangSizeCounts.get(x);
						double AllFrac = (double) AllVal / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) ;
						double RoundedAllFrac =(double)Math.round(AllFrac * 1000) / 10;
						double CoreFrac = (double) CoreVal / (double) (NonLargeCoreGeneCount);
						double RoundedCoreFrac = (double) Math.round(CoreFrac * 1000) / 10;
						
						//print output
						String str = "Size " + x + ": All: " + AllVal 
								+ "/" + (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " (" + String.valueOf(RoundedAllFrac) 
								+ "%), Core: " + CoreVal + "/" + (NonLargeCoreGeneCount)
								+ " (" +String.valueOf(RoundedCoreFrac) + "%).\n";
						
						if (x > 2){
							
							//write to file
							bw.write(str);
							bw.flush();
							
							//counters
							ThreeOrMoreAll = ThreeOrMoreAll + AllVal;
							ThreeOrMoreCore = ThreeOrMoreCore + CoreVal;
							
						} else {
							NotInGeneGangAll = NotInGeneGangAll + AllVal;
							NotInGeneGangCore = NotInGeneGangCore + CoreVal;
						}
					}
				}
				
				//stats for last line
				double AllFrac = (double) ThreeOrMoreAll / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) ;
				double RoundedAllFrac =(double)Math.round(AllFrac * 1000) / 10;
				double CoreFrac = (double) ThreeOrMoreCore / (double) (NonLargeCoreGeneCount);
				double RoundedCoreFrac = (double) Math.round(CoreFrac * 1000) / 10;
				
				//write last line
				String str = "Total: All: " + ThreeOrMoreAll 
						+ "/" + (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " (" + String.valueOf(RoundedAllFrac) 
						+ "%), Core: " + ThreeOrMoreCore + "/" + (NonLargeCoreGeneCount)
						+ " (" +String.valueOf(RoundedCoreFrac) + "%).\n\n";
				
				bw.write(str);
				bw.flush();
				
				/*
				 * GROUP 2 : Gang Size excludes large clusters
				 */
				
				//header
				String header2 = "Gang size excludes large clusters:\n";
				bw.write(header2);
				bw.flush();
			
				//3 or more large clusters, not in a gene gang
				ThreeOrMoreAll = 0;
				ThreeOrMoreCore = 0;
				NotInGeneGangAll = 0;
				NotInGeneGangCore = 0;
				
				Mappings = GGS.NonLargeGangSizeCounts;
				for (int x = 100; x >= 0; x--){
					if (Mappings.get(x) != null){
						
						//Retrieve / Compute values
						int AllVal = Mappings.get(x);
						int CoreVal = GGS.NonLargeCoreGangSizeCounts.get(x);
						AllFrac = (double) AllVal / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) ;
						RoundedAllFrac =(double)Math.round(AllFrac * 1000) / 10;
						CoreFrac = (double) CoreVal / (double) (NonLargeCoreGeneCount);
						RoundedCoreFrac = (double) Math.round(CoreFrac * 1000) / 10;
						
						//print output
						str = "Size " + x + ": All: " + AllVal 
								+ "/" + (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " (" + String.valueOf(RoundedAllFrac) 
								+ "%), Core: " + CoreVal + "/" + (NonLargeCoreGeneCount)
								+ " (" +String.valueOf(RoundedCoreFrac) + "%).\n";
						
						if (x > 2){
							
							//write to file
							bw.write(str);
							bw.flush();
							
							//counters
							ThreeOrMoreAll = ThreeOrMoreAll + AllVal;
							ThreeOrMoreCore = ThreeOrMoreCore + CoreVal;
							
						} else {
							NotInGeneGangAll = NotInGeneGangAll + AllVal;
							NotInGeneGangCore = NotInGeneGangCore + CoreVal;
						}
					}
				}
				
				//stats for last line
				AllFrac = (double) ThreeOrMoreAll / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) ;
				RoundedAllFrac =(double)Math.round(AllFrac * 1000) / 10;
				CoreFrac = (double) ThreeOrMoreCore / (double) (NonLargeCoreGeneCount);
				RoundedCoreFrac = (double) Math.round(CoreFrac * 1000) / 10;
				
				//write last line
				str = "Total: All: " + ThreeOrMoreAll 
						+ "/" + (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " (" + String.valueOf(RoundedAllFrac) 
						+ "%), Core: " + ThreeOrMoreCore + "/" + (NonLargeCoreGeneCount)
						+ " (" +String.valueOf(RoundedCoreFrac) + "%).\n\n";
				
				bw.write(str);
				bw.flush();
				
			}
			
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
		
	}
	
	//Export Gene Gang Values - not an AGG :-(
	public static void ExportGeneGangStats(String OutputFileName){
	
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(OutputFileName));
			String Line = null;
			for (double Ruliness : RulinessGGCountMapping.keySet()){

				Line = "Ruliness = " + String.valueOf(Ruliness) + "\n";
				bw.write(Line);
				bw.flush();
				
				GeneGangStats GGS = RulinessGGCountMapping.get(Ruliness);
				
				String header1 = "Gang size includes large clusters:\n";
				bw.write(header1);
				
				//3 or more large clusters, not in a gene gang
				int ThreeOrMoreSize = 0;
				int NotInGeneGangSize = 0;
				
				LinkedHashMap<Integer,Integer> Mappings = GGS.GangSizeCounts;
				for (int x = 100; x >= 0; x--){
					if (Mappings.get(x) != null){
						int Val = Mappings.get(x);
						
						double Frac = (double) Val / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) ;
						Frac = 100 * Frac;
						double RoundedFrac =(double)Math.round(Frac * 10) / 10;
						
						String str = "Size " + x + ": " + Mappings.get(x) 
								+ "/" + (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " or " + String.valueOf(RoundedFrac) 
								+ "% of non-large genes.\n";
						
						if (x > 1){
							bw.write(str);
							bw.flush();
						} else {
							NotInGeneGangSize = NotInGeneGangSize + Mappings.get(x);
						}
						
						//3 or more clusters
						if (x >= 3){
							ThreeOrMoreSize = ThreeOrMoreSize + Mappings.get(x);
						}
					}
				}

				//Three or more
				double Frac3 = (double) ThreeOrMoreSize / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) ;
				Frac3 = 100 * Frac3;
				double RoundedFrac3 =(double)Math.round(Frac3 * 10) / 10;
				
				String str3 = "Three or more: " + ThreeOrMoreSize 
						+ "/" + (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " or " + String.valueOf(RoundedFrac3)
						+ "% of non-large genes.\n";
			
				bw.write(str3);
				bw.flush();
				
				// Not in a gene gnag
				double FracNon = (double) NotInGeneGangSize / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount);
				FracNon = 100 * FracNon;
				double RoundedFracNon = (double) Math.round(FracNon * 10) / 10;
				
				String strnon = "Not in a gene gang: " + NotInGeneGangSize + "/"
						+ (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " or " + String.valueOf(RoundedFracNon)
						+ "% of non-large genes.\n\n";
				
				bw.write(strnon);
				bw.flush();

				/*
				 * SECOND SET: gang size calculated based on num of non-large clusters:
				 */

				//header
				String header2 = "Gang size does not include non-large clusters:\n";
				bw.write(header2);
				
				//re-set counters
				ThreeOrMoreSize = 0;
				NotInGeneGangSize = 0;
				
				Mappings = GGS.NonLargeGangSizeCounts;
				for (int x = 100; x >= 0; x--){
					if (Mappings.get(x) != null){
						int Val = Mappings.get(x);
						
						double Frac = (double) Val / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) ;
						Frac = 100 * Frac;
						double RoundedFrac =(double)Math.round(Frac * 10) / 10;
						
						String str = "Size " + x + ": " + Mappings.get(x) 
								+ "/" + (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " or " + String.valueOf(RoundedFrac) 
								+ "% of non-large genes.\n";
						
						if (x > 1){
							bw.write(str);
							bw.flush();
						} else{
							NotInGeneGangSize = NotInGeneGangSize + Mappings.get(x);
						}

						
						//3 or more clusters
						if (x >= 3){
							ThreeOrMoreSize = ThreeOrMoreSize + Mappings.get(x);
						}
					}
				}
				
				//Three or more (second option)
				Frac3 = (double) ThreeOrMoreSize / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) ;
				Frac3 = 100 * Frac3;
				RoundedFrac3 =(double)Math.round(Frac3 * 10) / 10;
				
				str3 = "Three or more: " + ThreeOrMoreSize 
						+ "/" + (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " or " + String.valueOf(RoundedFrac3)
						+ "% of non-large genes.\n";
			
				bw.write(str3);
				bw.flush();
				
				// Not in a gene gnag
				FracNon = (double) NotInGeneGangSize / (double) (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount);
				FracNon = 100 * FracNon;
				RoundedFracNon = (double) Math.round(FracNon * 10) / 10;
				
				strnon = "Not in a gene gang: " + NotInGeneGangSize + "/"
						+ (NonLargeCoreGeneCount+NonLargeNonCoreGeneCount) + " or " + String.valueOf(RoundedFracNon)
						+ "% of non-large genes.\n";
				
				bw.write(strnon);
				bw.flush();
				
				bw.write("------------------------------------\n\n");
				bw.flush();

			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export gene gangs with confidence values included
	public static void ExportPBCV1SummaryTablev2(AmalgamatedGeneGang GeneGangs, String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));			
			
			//Export Header
			String Header = "Gang\tRuliness\tPBCV-1_ID\tAnnotations\tCluster_ID\tExpression/Virion-Association\n";
			String Underline = "-------------------------------------------------------------" +
					"----------------------------------------------------------\n";
			bw.write(Header);
			bw.write(Underline);
			bw.flush();

			//Initialize list of lists
			LinkedList<LinkedList<GenomicFeature>> PBCV1GeneCollection = new LinkedList<LinkedList<GenomicFeature>>();
			
			//large groups only
			LinkedList<LinkedList<GenomicFeature>> PBCV1_LargeGG = new LinkedList<LinkedList<GenomicFeature>>();
			
			//mod 10/20/2013
			for (Integer x : GeneGangs.GenomicFeaturesByGangNumber.keySet()){
				
				//retrieve features
				LinkedList<GenomicFeature> Features = GeneGangs.GenomicFeaturesByGangNumber.get(x);
				LinkedList<GenomicFeature> PBCV1Genes = new LinkedList<GenomicFeature>();
				
				//retrieve PBCV-1 genes
				for (GenomicFeature GF : Features){
				if (GF.Contig.equals("PBCV-1")){
						PBCV1Genes.add(GF);
					}
				}
				
				//sort
				Collections.sort(PBCV1Genes, new GFListSorter());	
				
				//add to list
				PBCV1GeneCollection.add(PBCV1Genes);
				
				// if large enough, add to the large list
				if (PBCV1Genes.size() > 2){
					PBCV1_LargeGG.add(PBCV1Genes);
				}
				
			}
			
			//sort lists
			Collections.sort(PBCV1GeneCollection, new LLSorter());
			Collections.sort(PBCV1_LargeGG, new LLSorter());
			
			//prepare to return this value
			GeneGangs.PBCV1_LargeGG = PBCV1_LargeGG;
			
			//Variables for file writing
			int GangCounter = 0;
			String Entry = "";
			
			//write sorted list to file
			for (LinkedList<GenomicFeature> PBCV1Genes : PBCV1GeneCollection){
				
				//increment counter
				GangCounter++;
				
				for (GenomicFeature GF : PBCV1Genes){
					Entry = String.valueOf(GangCounter) + "\t" 
						+ String.valueOf(GeneGangs.MaximumUnrulinessMap.get(GF)) + "\t"
						+ JCEReformatGeneID(GF) + "\t" 
						+ ClusterAnnotations.get(GF.ClusterNum) + "\t"
						+ String.valueOf(GF.ClusterNum) + "\t\t\t\n";
					bw.write(Entry);
					bw.flush();
				}
				
				//new line between entries
				bw.write("\n");
				bw.flush();	
				
			}			
			
			//close file stream
			bw.close();
			
		} catch (Exception ex) {
			ex.printStackTrace();
		}

	}
	
	//Export according to new, post-amalgamated unruliness set
	public static void Step3Exports(AmalgamatedGeneGang AGGr,String FileName){
		
		//Files
		String GeneGangContextSetFile = GeneGangContextSetDir + "/" + FileName;
		String GeneGangsAsIDFile = GGListDir + "/" + FileName;
		String AmalgFile = AmalgDir + "/" + FileName;
		
		//Context Set
		ExportGeneGangContextSets(AGGr,GeneGangContextSetFile);
		
		//Gene List
		ExportGeneGangGeneIDs(AGGr, GeneGangsAsIDFile);
		
		//Cluster List
		ExportAmalgamatedGeneGangs2(AGGr,AmalgFile);
	}
	
	// ====== Main method ====== //

	//OLD - run back in 2013/2014
	//main method
	public static void main(String[] args) {

		//Load + Prepare all data
		LoadAllData();
				
//		//Initialize a table, to store output
//		LinkedList<AmalgamatedTableRow> DistTable = new LinkedList<AmalgamatedTableRow>();
		
		//parameters
		
		int Rmin = 7;
		int Rmax = 7;
		int Rinc = 1;
		
		UScale = 0.025;

		int Umin = (int) Math.round(0.82500/UScale);
		int Umax = (int) Math.round(0.82500/UScale);
		
//		//store values in this matrix
//		AmalgamatedTableRow[][] ATRMatrix = new AmalgamatedTableRow[(Umax-Umin) + 1][((Rmax-Rmin)/Rinc)+1];
		
		//initialize mappings
		LinkedHashMap<Integer, Double> Comparisons = new LinkedHashMap<Integer, Double>();
		
		//Range values
		for (int i = Rmin; i <= Rmax; i = i + Rinc){
			
			//Range value
			int Range = i*1000;
						
			//Initialize statistics hash map
			RulinessGGCountMapping = new LinkedHashMap<Double, GeneGangStats>();
			LinkedList<AmalgamatedGeneGang> PreAmalgUnrulinessList = new LinkedList<AmalgamatedGeneGang>();
			
			//pre-amalgamation unruliness
			for (int j = Umin; j <= Umax; j++){
				
				//turn integer to double + remove arithmetic errors
				double U = UScale* (double) j;		
				U = (double)Math.round(U * 1000) / 1000;
				double AU = 1.000 - U;
				AU = (double)Math.round(AU * 1000) / 1000;

				//File name
				String FileName = String.valueOf(i) +"K_" + String.valueOf(AU) +".txt";
				
				//compute gene gangs, add summary statistics to table
				AmalgamatedTableRow ATR = MultipleGeneGangs(Range, U, AU, FileName);
				PreAmalgUnrulinessList.add(0,ATR.AGG);
//				DistTable.add(ATR);
				
				//write result to matrix
				//ATRMatrix[Umax-j][(Rmax-i)/Rinc] = ATR;
				
			}
			
			//Display mappings
			String GeneGangStatsFile = StatisticsDir + "/PreAmalgStats_" + String.valueOf(i) + "K.txt";
			ExportGeneGangStats(GeneGangStatsFile);
			
			//Compare large gangs
			//Double GangDiff = CompareLargePBCV1Amalgamations(Rows.get(Umin).AGG, Rows.get(Umax).AGG);
			
			//Comparisons.put(Range,GangDiff);
			
			//find set of gene gangs associated before the breaking point.
			AmalgamatedGeneGang AGG = PreAmalgamatedBreakingPoint(PreAmalgUnrulinessList);
			if (AGG != null){
				System.out.println("last before breaking: " + AGG.PreAmalgamatedUnruliness);
			} else {
				System.out.println("All broken at this range value.");
			}
			
			//Re-analyze this breaking point
			DeterminePostAmalgamatedUnruliness(AGG);
			LinkedList<Double> UnrulinessList = AGG.PostAmalgamatedUnrulinessRange;
			
			//Export PBCV-1 data
			String PBCV1Table = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/PBCV1SummaryTable.txt";
			ExportPBCV1SummaryTablev2(AGG,PBCV1Table);
			
			//values for storage
			LinkedHashMap<Double, GeneGangStats> UnrulinessMapping =
					new LinkedHashMap<Double, GeneGangStats>();
			
			//iterate through list
			for (Double d : UnrulinessList){
				
				//compute values
				AmalgamatedGeneGang AGGr = RecomputeBasedOnUnruliness(AGG,d);
				String FileName = String.valueOf(i) +"K_" + String.valueOf(d) +".txt";
				
				//Stats + Store values
				GeneGangStats GGS = PostAmalgamationStats(AGGr);
				UnrulinessMapping.put(d, GGS);
			
				//Exports
				Step3Exports(AGGr,FileName);
				
				exportGeneGangSizeTable(AGGr, d);
				exportGeneGangRangesTable(AGGr, d);
			}
			
			//Write statistics
			String TestFile = StatisticsDir +"/PostAmalgamationStats.txt";
			ExportGeneGangStatsv2(TestFile,UnrulinessMapping);
			
			//phylotree mapping
			PhyloTrees PT = new PhyloTrees(AGG);
			
//			//debugging - check clusters + protein annotations
//			for (ClusterData CD : ClusterDirectory){
//				if (CD.isCore && !CD.isLargeCluster){
//					PT.ClusterProtData(CD.ClusterNum);
//				}
//			}
			
			//export sequences for alignments
			for (Integer x : AGG.GenomicFeaturesByGangNumber.keySet()){
				if (AGG.ClustersFeatured.get(x).size() >= 3){
						
					String FileName = ProtSeqDir 
							+ "/Gang" + String.valueOf(x) + "_"
							+ String.valueOf(AGG.ClustersFeatured.get(x)).replace(" ","_")
							+ ".fasta";
					
					LinkedHashMap<Integer,LinkedList<GenomicFeature>> 
						ClusterData = PT.ExportFileForAlignment(x,FileName);
					
					//send to export
					for (Integer ClustNum : ClusterData.keySet()){
						
						//file name
						String ClusterFileName = SingleClusterDir + "/Cluster" + String.valueOf(ClustNum) + ".fasta";
						
						//retrieve features
						LinkedList<GenomicFeature> FeatureList = ClusterData.get(ClustNum);
						
						//export
						PT.ExportBunch(FeatureList, ClusterFileName);
						
					}
					
				}
			}
		}
	
//		System.out.println("Comparisons:");
//		for (Integer x : Comparisons.keySet()){
//			System.out.println(x +"\t" + Comparisons.get(x));
//		}
		
//		//process matrix
//		ATRMatrix = ProcessMatrix(ATRMatrix);
//		
//		//Export table
//		//String TableFile = BaseDir + "/OverallSummaryTable.txt";
//		//ExportSummaryTable(DistTable,TableFile);
//		
//		//create heat map
//		String DissMatrixFile = BaseDir + "/DissimilarityMap.txt";
//		ExportDissimilarityMap(ATRMatrix, DissMatrixFile);
//		
//		//create dissimilarity map with averages
//		String MapFile = BaseDir + "/LocalVarianceMatrix.txt";
//		ExportMatrixForHeatMap(ATRMatrix, MapFile);
		
		//PrintATRMatrix(ATRMatrix);
		//ExportParameterHeatMap
		
		//final output message
		System.out.println("All Processes Successfully Completed!");

	}
	
	// ==== Deprecated ===== //
	
	//Export dissimilarity map
	public static void ExportMatrixForHeatMap(AmalgamatedTableRow[][] ATRMatrix, String DissTableFile){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(DissTableFile));
			for (int i = 0; i < ATRMatrix.length; i++){
				String Row = "";
				for (int j = 0; j <ATRMatrix[i].length; j++){
					Row = Row + ATRMatrix[i][j].AGG.AverageDiss + "\t";
				}
				bw.write(Row);
				bw.write("\n");
				bw.flush();
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}

	//Export a heat map.
	public static void ExportDissimilarityMap(AmalgamatedTableRow[][] ATRMatrix, String DissMapFile){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(DissMapFile));

			for (int i = 0; i < ATRMatrix.length; i++){
				String MainRow = "";
				String SubRow = "";
				for (int j = 0; j < ATRMatrix[i].length; j++){
					
					int RangeScaled = ATRMatrix[i][j].Range / 1000;
					String RangeNum = String.valueOf(RangeScaled) + "K";
					
					MainRow = MainRow + "(" + RangeNum + "," + String.valueOf(ATRMatrix[i][j].Ruliness) + ")" + "\t";

					//add below diss
					if (i < ATRMatrix.length - 1){
						SubRow = SubRow + ATRMatrix[i][j].AGG.Below + "\t";
					}
					
					//add right-ward diss
					if (j < ATRMatrix[i].length - 1){
						MainRow = MainRow + ATRMatrix[i][j].AGG.Right + "\t";
						SubRow = SubRow + "\t\t";
					}
					
				}
				MainRow = MainRow + "\n";
				SubRow = SubRow + "\n";
				bw.write(MainRow);
				bw.write(SubRow);
				bw.flush();
			}
			
			bw.close();
			
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}

	//Export a summary output table.
	public static void ExportSummaryTable(LinkedList<AmalgamatedTableRow> DistTable, String TableFile){

		try {
			//new writer
			BufferedWriter bw = new BufferedWriter(new FileWriter(TableFile));
			
			//this value refers to all gene gangs with more than 10 members.
			int MC = 11;
			
			//Headers
			String Header2 = "Rng/Unr\t>10";
			for (int i = MC-1; i >= 1; i--){
				Header2 = Header2  + "\t" + String.valueOf(i);
			}
			Header2 = Header2 + "\tTot\n";
			bw.write(Header2);
			bw.flush();
			
			//Table Content
			for (AmalgamatedTableRow ATR : DistTable){
				int NewRange = (ATR.Range / 1000);
				double AU = (double)Math.round((1.0-ATR.Unruliness)*1000) / 1000;
				String Row = String.valueOf(NewRange) + "K/" + String.valueOf(AU) + "\t";
				int Total = 0;
				for (int i = MC; i>= 1; i--){
					String Val = "0";
					if (ATR.Counts.get(i) != null){
						Val = String.valueOf(ATR.Counts.get(i));
						Total = Total + ATR.Counts.get(i);
					}
					Row = Row + Val + "\t";
				}
				Row = Row + String.valueOf(Total) + "\n";
				bw.write(Row);
				bw.flush();
			}
			bw.close();
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	
	
	//Created on 2016-07-28, to generate for a given unruliness, a table of gene size, organism, and the clusters in each gene gang.
	public static void exportGeneGangSizeTable(AmalgamatedGeneGang AGG, Double unruliness){
				
		String tableDir = "/Users/phillipseitzer/UCDavis/GeneGangs-2016-Revival/2016-07-28-gene-gangs-TEST/output/gang-stats/" + unruliness + ".txt";
		
		
		try {
			String tableFile = tableDir;
			BufferedWriter bw = new BufferedWriter(new FileWriter(tableFile));
			
			LinkedList<String> orgNames = new LinkedList<String>();
			
			LinkedHashMap<Integer, HashSet<Integer>> clusterNums = new LinkedHashMap<Integer, HashSet<Integer>>();
			
			LinkedHashMap<Integer, Float> averages = new LinkedHashMap<Integer, Float>();
			
			//			org name,			   gang num,  size
			LinkedHashMap<String, LinkedHashMap<Integer, Integer>> tableData = new LinkedHashMap<String, LinkedHashMap<Integer, Integer>>();
			
			LinkedHashMap<Integer, LinkedHashMap<String, LinkedList<GenomicFeature>>> gangs = AGG.GeneGangsByOrg;
			for (Entry<Integer, LinkedHashMap<String, LinkedList<GenomicFeature>>> entry : gangs.entrySet()){
				
				int gangNum = entry.getKey();
				
				if (AGG.ClustersFeatured.get(gangNum).size() < 3){
					continue;
				}
				
				LinkedHashMap<String, LinkedList<GenomicFeature>> gangData = entry.getValue();
				
				int sum = 0;
								
				for (Entry<String, LinkedList<GenomicFeature>> orgData : gangData.entrySet()){
					
					HashSet<Integer> clustNums = new HashSet<Integer>();
					
					String name = orgData.getKey();
					
					if (!orgNames.contains(name)){
						orgNames.add(name);
					}
					
					LinkedList<GenomicFeature> features = orgData.getValue();
					
					int gangSize = 0;
					
					if (features.size() >= 2){
						gangSize = features.get(features.size()-1).StopPos - features.get(0).StartPos;
					} else if (features.size() == 1){
						gangSize = features.get(0).StopPos - features.get(0).StartPos;
					}
					
					for (GenomicFeature gf : features){
						clustNums.add(gf.ClusterNum);
					}
					
					sum += gangSize;
					
					LinkedHashMap<Integer, Integer> orgGangData = new LinkedHashMap<Integer, Integer>();
					
					if (tableData.containsKey(name)){
						orgGangData = tableData.get(name);
					}
					
					orgGangData.put(gangNum, gangSize);
					tableData.put(name, orgGangData);
					
					if (clusterNums.containsKey(gangNum)){
						clusterNums.get(gangNum).addAll(clustNums);
					} else {
						clusterNums.put(gangNum, clustNums);
					}
				}
				
				averages.put(gangNum, (float) sum / (float) orgNames.size());
			}
			
			Collections.sort(orgNames);
			
			LinkedList<Integer> validGangNums = new LinkedList<Integer>();
				
			StringBuilder sb = new StringBuilder();
			
			for (String orgName : orgNames){
								
				sb.append(orgName);
				sb.append("\t");
				
				LinkedHashMap<Integer, Integer> stats = tableData.get(orgName);
				
				for (int x = 0; x < 100; x++){
					if (stats.containsKey(x)){
						if (!validGangNums.contains(x)){
							validGangNums.add(x);
						}
						sb.append("\t");
						sb.append(String.valueOf(stats.get(x)));
					}
				}
				
				sb.append("\n");
			}
			
			bw.write("\t\t");
			for (Integer x : validGangNums){
				HashSet<Integer> clusters = clusterNums.get(x);
				LinkedList<Integer> sortedClusters = new LinkedList<Integer>(clusters);
				Collections.sort(sortedClusters);
				StringBuilder sb2 = new StringBuilder();
				sb2.append("[");
				for (int i = 0 ; i < sortedClusters.size(); i++){
					if (i > 0){
						sb2.append(",");
					}
					sb2.append(String.valueOf(sortedClusters.get(i)));
				}
				sb2.append("]\t");
				
				bw.write(sb2.toString());
			}
			
			bw.write("\n");
			
			bw.write(sb.toString());
			bw.flush();
			
			bw.append("\n");
			
			StringBuilder sb3 = new StringBuilder();
			sb3.append("average");
			sb3.append("\t");
			
			for (Integer x : validGangNums){
				sb3.append("\t");
				sb3.append(String.valueOf(averages.get(x)));
			}
			
			bw.write(sb3.toString());
			bw.flush();
			
			bw.close();
			
		} catch (IOException e){
			e.printStackTrace();
		} 
	}

	public static void exportGeneGangRangesTable(AmalgamatedGeneGang AGG, Double unruliness){
		
		String dataFile = "/Users/phillipseitzer/UCDavis/GeneGangs-2016-Revival/2016-08-5-gene-gang-lengths/" + unruliness + ".txt";
		
		
		try {
			String tableFile = dataFile;
			BufferedWriter bw = new BufferedWriter(new FileWriter(tableFile));
			
			LinkedList<String> orgNames = new LinkedList<String>();
			
			LinkedHashMap<Integer, HashSet<Integer>> clusterNums = new LinkedHashMap<Integer, HashSet<Integer>>();
			
			LinkedHashMap<Integer, Float> averages = new LinkedHashMap<Integer, Float>();
			
			//			org name,			   gang num,  String
			LinkedHashMap<String, LinkedHashMap<Integer, String>> tableData = new LinkedHashMap<String, LinkedHashMap<Integer, String>>();
			
			LinkedHashMap<Integer, LinkedHashMap<String, LinkedList<GenomicFeature>>> gangs = AGG.GeneGangsByOrg;
			for (Entry<Integer, LinkedHashMap<String, LinkedList<GenomicFeature>>> entry : gangs.entrySet()){
				
				int gangNum = entry.getKey();
				
				if (AGG.ClustersFeatured.get(gangNum).size() < 3){
					continue;
				}
				
				LinkedHashMap<String, LinkedList<GenomicFeature>> gangData = entry.getValue();
				
				int sum = 0;
								
				for (Entry<String, LinkedList<GenomicFeature>> orgData : gangData.entrySet()){
					
					HashSet<Integer> clustNums = new HashSet<Integer>();
					
					String name = orgData.getKey();
					
					if (!orgNames.contains(name)){
						orgNames.add(name);
					}
					
					LinkedList<GenomicFeature> features = orgData.getValue();
					
					int gangSize = 0;
					
					int gangStart = Integer.MAX_VALUE;
					int gangStop = 0;
					
					for (GenomicFeature gf : features){
						if (gf.StartPos < gangStart){
							gangStart = gf.StartPos;
						}
						if (gf.StopPos < gangStart){
							gangStart = gf.StopPos;
						}
						
						if (gf.StartPos > gangStop){
							gangStop = gf.StartPos;
						}
						
						if (gf.StopPos > gangStop){
							gangStop = gf.StopPos;
						}
					}
					
					String gangString = String.valueOf(gangStart)+ "-"+ String.valueOf(gangStop);
					
					if (features.size() >= 2){
						gangSize = features.get(features.size()-1).StopPos - features.get(0).StartPos;
					} else if (features.size() == 1){
						gangSize = features.get(0).StopPos - features.get(0).StartPos;
					}
					
					for (GenomicFeature gf : features){
						clustNums.add(gf.ClusterNum);
					}
					
					sum += gangSize;
					
					LinkedHashMap<Integer, String> orgGangDuration = new LinkedHashMap<Integer, String>();
					
					if (tableData.containsKey(name)){
						orgGangDuration = tableData.get(name);
					}
					
					orgGangDuration.put(gangNum, gangString);
					tableData.put(name, orgGangDuration);
					
					if (clusterNums.containsKey(gangNum)){
						clusterNums.get(gangNum).addAll(clustNums);
					} else {
						clusterNums.put(gangNum, clustNums);
					}
				}
				
				averages.put(gangNum, (float) sum / (float) orgNames.size());
			}
			
			Collections.sort(orgNames);
			
			LinkedList<Integer> validGangNums = new LinkedList<Integer>();
				
			StringBuilder sb = new StringBuilder();
			
			for (String orgName : orgNames){
								
				sb.append(orgName);
				sb.append("\t");
				
				LinkedHashMap<Integer, String> stats = tableData.get(orgName);
				
				for (int x = 0; x < 100; x++){
					if (stats.containsKey(x)){
						if (!validGangNums.contains(x)){
							validGangNums.add(x);
						}
						sb.append("\t");
						sb.append(stats.get(x));
					}
				}
				
				sb.append("\n");
			}
			
			bw.write("\t\t");
			for (Integer x : validGangNums){
				HashSet<Integer> clusters = clusterNums.get(x);
				LinkedList<Integer> sortedClusters = new LinkedList<Integer>(clusters);
				Collections.sort(sortedClusters);
				StringBuilder sb2 = new StringBuilder();
				sb2.append("[");
				for (int i = 0 ; i < sortedClusters.size(); i++){
					if (i > 0){
						sb2.append(",");
					}
					sb2.append(String.valueOf(sortedClusters.get(i)));
				}
				sb2.append("]\t");
				
				bw.write(sb2.toString());
			}
			
			bw.write("\n");
			
			bw.write(sb.toString());
			bw.flush();
			
			bw.append("\n");
			
			StringBuilder sb3 = new StringBuilder();
			sb3.append("average");
			sb3.append("\t");
			
			for (Integer x : validGangNums){
				sb3.append("\t");
				sb3.append(String.valueOf(averages.get(x)));
			}
			
			//bw.write(sb3.toString());
			bw.flush();
			
			bw.close();
			
		} catch (IOException e){
			e.printStackTrace();
		} 
	}
	
	//Process matrix with various parameters
	public static AmalgamatedTableRow[][] ProcessMatrix(AmalgamatedTableRow[][] ATRMatrix){
		
		//determine all pairwise dissimilarities
		for (int i = 0; i < ATRMatrix.length; i++){
			for (int j = 0; j < ATRMatrix[i].length; j++){
				
				//Below
				if (i < ATRMatrix.length - 1){
					double ValBelow = CompareAmalgamations(ATRMatrix[i][j].AGG,ATRMatrix[i+1][j].AGG);
					ATRMatrix[i][j].AGG.Below = (double)Math.round(ValBelow * 100) / 100;
				}
				
				//To the right
				if (j < ATRMatrix[i].length - 1){
					double ValRight = CompareAmalgamations(ATRMatrix[i][j].AGG,ATRMatrix[i][j+1].AGG);
					ATRMatrix[i][j].AGG.Right = (double)Math.round(ValRight * 100) / 100;
				}
			}
		}
		
		
		double minAverageDiss = 1.0;
		int Range = 0;
		double Ruliness = 0.0;
		
		//identify average neighbor-dissimilarity
		for (int i = 0; i < ATRMatrix.length; i++){
			for (int j = 0; j < ATRMatrix[i].length; j++){
				
				//initialize
				double OverallAverage = 0.0;
				double NumEntries = 0;
				
				//not in first row
				if (i != 0){
					OverallAverage = OverallAverage + ATRMatrix[i-1][j].AGG.Below;
					NumEntries++;
				}
				
				//not in last row
				if (i != ATRMatrix.length -1){
					OverallAverage = OverallAverage + ATRMatrix[i][j].AGG.Below;
					NumEntries++;
				}
				
				
				//not in first column
				if (j != 0){
					OverallAverage = OverallAverage + ATRMatrix[i][j-1].AGG.Right;
					NumEntries++;
				}

				//not in last column
				if (j != ATRMatrix[i].length -1){
					OverallAverage = OverallAverage + ATRMatrix[i][j].AGG.Right;
					NumEntries++;
				}
				
				double avg = (double) Math.round(100*OverallAverage/NumEntries)/100 ;
				
				if (avg < minAverageDiss){
					minAverageDiss = avg;
					Range = ATRMatrix[i][j].Range;
					Ruliness = ATRMatrix[i][j].Ruliness;
					
				}
				//write average
				ATRMatrix[i][j].AGG.AverageDiss = avg;
						
			}
		}
		
		//min average
		System.out.println("Minimum Average Diss: " 
				+ minAverageDiss + " from (" + Range + "," + Ruliness + ")");
		
		return ATRMatrix;
	}
}
