package FileFormatting;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedList;

import FileFormatting.LoadData.GenomicFeature;
import FileFormatting.MultipleMemberGeneGangs.AmalgamatedGeneGang;

public class PhyloTrees extends LoadData{

	/**
	 * @param args
	 */
	
	//Fields
	//Data
	public static AmalgamatedGeneGang GG;
	
	//classes
	public class ClusterNumSorter implements Comparator<GenomicFeature>{

		@Override
		public int compare(GenomicFeature o1, GenomicFeature o2) {
			if (o1.ClusterNum != o2.ClusterNum){
				return o1.ClusterNum-o2.ClusterNum;		//Primary: cluster number
			} else {
				return o1.Size - o2.Size; 				//Secondary: size
			}

		}
		
	}
	
	//Constructor
	public PhyloTrees(AmalgamatedGeneGang AGG){
		GG = AGG;
	}
	
	//silly test method
	public void PhyloTreeTest(){
		
		LinkedList<GenomicFeature> Features = null;
		for (int x = 0; x < 1000; x++){
			if (GG.GenomicFeaturesByGangNumber.get(x) != null){
				Features = GG.GenomicFeaturesByGangNumber.get(x);
				break;
			}
		}
		
		for (GenomicFeature GF : Features){
			LinkedList<String> L = ProteinSequence(GF);
			for (String s : L){
				System.out.println(s);
			}
			//System.out.println(L.get(0));
		}
		
	}
	
	// ======== Sequence 2 Alignment Methods ======== //
	
	//Given a list of genomic features, return a single aligned protein sequence.
	public LinkedList<String> OrderedGang2Seq(LinkedList<GenomicFeature> OrderedFeatureList, String Header){
		
		//Initialize
		LinkedList<String> AmalgamatedProteinSequences = new LinkedList<String>();
		
		//write header
		AmalgamatedProteinSequences.add(Header);
		
		//begin writing sequence
		String AmalgamatedSeqs = "";
		
		//Retrieve sequences
		for (GenomicFeature GF : OrderedFeatureList){
			LinkedList<String> Seq = ProteinSequence(GF);
			AmalgamatedSeqs = AmalgamatedSeqs + Seq.get(1);
		}
		
		//connect to header
		AmalgamatedSeqs = AmalgamatedSeqs + "\n";
		AmalgamatedProteinSequences.add(AmalgamatedSeqs);
		
		//output
		return AmalgamatedProteinSequences;
		
	}
	
	//Hash map -> export
	public void ExportHash(LinkedHashMap<String,LinkedList<GenomicFeature>> FeatureMapping, String FileName){
		try {
			for (String s : FeatureMapping.keySet()){
				
				
			}
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Single cluster - check data
	public void ClusterProtData(Integer ClusterNum){
		int ProtCounter = 0;
		int InstanceCounter = 0;
		boolean FoundProt = false;
		for (String s : Chloroviruses.keySet()){
			FoundProt = false;
			LinkedList<GenomicFeature> Features = Chloroviruses.get(s);
			for (GenomicFeature GF : Features){
				if (GF.ClusterNum == ClusterNum){
					InstanceCounter++;
					LinkedList<String> ProtSeq = ProteinSequence(GF);
					if (ProtSeq.size() > 0){
						FoundProt = true;
						ProtCounter++;
					}
				}
			}
			if (!FoundProt){
				System.out.println("Miss: " + s);
			}
		}
		
		System.out.println("Cluster " + ClusterNum 
				+ ": Instances:" + InstanceCounter + ", Prots: " + ProtCounter);
	}
	
	//Single cluster - export
	public void ExportBunch(LinkedList<GenomicFeature> FeatureList, String FileName){
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
			for (GenomicFeature GF : FeatureList){
				String Contig = GF.Contig;
				if (Contig.equals("PBCV")){
					Contig = "PBCV1";
				}
				String Header = ">" + Contig + "\n";
				String Seq = GF.ProteinSequence + "\n\n";
				bw.write(Header);
				bw.write(Seq);
				bw.flush();
			}
			bw.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Methods
	
	//Massive export - for exporting gene gang sequence trees
	public LinkedHashMap<Integer,LinkedList<GenomicFeature>> ExportFileForAlignment(Integer GGNumber, String FileName){
		/*
		 * This method exports a file containing the nucleotide sequences of all relevant genes
		 * mashed together into a concatenated super-sequence.
		 * 
		 * This sequence may then be aligned using a multiple-alignment program, and
		 * fed into a phylogenetic tree generation program.
		 * 
		 * Several considerations here:
		 * 
		 * (1) Gene order may change
		 * (2) Multiple paralogs exist, so need to map appropriate homologs to each other
		 * (3) Gene strandedness may change
		 * (4) Needs to contain only the maximum overlapping comparable parts - so, if a gene
		 *     gang contains a gene in one organism that is not found in another, the gene is excluded.
		 * 
		 * Others?
		 * 
		 * Algorithm:
		 * (1) Determine appropriate genes / orientations in each organism for relevant gene export
		 * (2) Concatenate appropriate sequences + export
		 */
		
		//Initialize output
		LinkedHashMap<Integer,LinkedList<GenomicFeature>> FinalClusters =
				new LinkedHashMap<Integer,LinkedList<GenomicFeature>>();
	
		/*
		 * DETERMINE GENES TO AMALGAMATE
		 */
		
		//(1) Retrieve protein sequences + associated with genomic elements
		
		LinkedList<GenomicFeature> AllFeatures = GG.GenomicFeaturesByGangNumber.get(GGNumber);
		
		for (GenomicFeature GF : AllFeatures){
			LinkedList<String> ProtSeq = ProteinSequence(GF);
			if (ProtSeq.size() > 0){
				GF.ProteinSequence = ProtSeq.get(1);
			}
		}
		
		//(1) Find minimal cluster set
		
		//Initialize a map of counts
		LinkedHashMap<Integer,Integer> AllMinClusterCounts = new LinkedHashMap<Integer,Integer>();
		boolean EmptyHash = true;
		
		//check all species-specific gang instances
		LinkedHashMap<String, LinkedList<GenomicFeature>> Instances = GG.GeneGangsByOrg.get(GGNumber);
		
		//A mapping to help discern difficult paralog cases.
		LinkedHashMap<Integer, LinkedList<GenomicFeature>> PreferredClusterVersions 
			= new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
		
		//re-organization of genomic features
		LinkedHashMap<String, LinkedHashMap<Integer,LinkedList<GenomicFeature>>> GangFeaturesHash
			= new LinkedHashMap<String,LinkedHashMap<Integer,LinkedList<GenomicFeature>>>();
		
		//check all organisms
		for (String s : Instances.keySet()){
			
			//Initialize a map of counts for this organism + map for re-mapped features
			LinkedHashMap<Integer,Integer> OrgClusterCounts = new LinkedHashMap<Integer,Integer>();
			LinkedHashMap<Integer, LinkedList<GenomicFeature>> GangMems = new LinkedHashMap<Integer, LinkedList<GenomicFeature>>();
			
			//retrieve organismal features
			LinkedList<GenomicFeature> Features = Instances.get(s);
					
			//increment hash
			int Val = 0;
			for (GenomicFeature GF : Features){
				
				//Determine counts
				Val = 0;
				if (OrgClusterCounts.get(GF.ClusterNum) != null){
					Val = OrgClusterCounts.get(GF.ClusterNum);
					
					//is this protein sequence available?
					if (GF.ProteinSequence != null){
						Val++;
					}
					
					OrgClusterCounts.put(GF.ClusterNum,Val);
				} else {
					
					//is this protein sequence available?
					if (GF.ProteinSequence != null){
						Val++;
					}
					
					OrgClusterCounts.put(GF.ClusterNum,Val);
				}
				
				//re-organize
				if (GF.ProteinSequence != null){
					if (GangMems.get(GF.ClusterNum) != null){
						LinkedList<GenomicFeature> GangFeatures = GangMems.get(GF.ClusterNum);
						GangFeatures.add(GF);
						GangMems.put(GF.ClusterNum, GangFeatures);
					} else{
						LinkedList<GenomicFeature> GangFeatures = new LinkedList<GenomicFeature>();
						GangFeatures.add(GF);
						GangMems.put(GF.ClusterNum,GangFeatures);
					}
				}
							
			}
			
			//update

			GangFeaturesHash.put(s, GangMems);
			
			//update the master
			if (EmptyHash){
				
				//First set - write all
				AllMinClusterCounts = OrgClusterCounts;
				
				//default - these genes are the preferred cluster versions
				for (GenomicFeature GF : Features){
					if (GF.ProteinSequence != null){
						LinkedList<GenomicFeature> ClustList;
						if (PreferredClusterVersions.get(GF.ClusterNum) != null){
							ClustList = PreferredClusterVersions.get(GF.ClusterNum);
						} else{
							ClustList = new LinkedList<GenomicFeature>();
						}
						ClustList.add(GF);
						PreferredClusterVersions.put(GF.ClusterNum, ClustList);
					}
				}
				
				//hash set not empty, any more
				EmptyHash = false;
				
			} else { //modify
				
				//for clusters found in both lists, pick smallest value
				for (Integer x : OrgClusterCounts.keySet()){
					if (AllMinClusterCounts.get(x) != null){
						if (OrgClusterCounts.get(x) < AllMinClusterCounts.get(x)){
							
							//update counts
							AllMinClusterCounts.put(x,OrgClusterCounts.get(x));
							
							//update sequence versions
							LinkedList<GenomicFeature> ClustList = new LinkedList<GenomicFeature>();
							for (GenomicFeature GF : Features){
								if (GF.ProteinSequence != null){
									if (GF.ClusterNum == x){
										ClustList.add(GF);
									}
								}
							}
							
							//replace old list with the new list
							PreferredClusterVersions.put(x, ClustList);
							
						}
					}
				}
				
				//Any entries not found in this org set are removed from the master set.
				for (Integer x : AllMinClusterCounts.keySet()){
					if (OrgClusterCounts.get(x) == null){
						AllMinClusterCounts.put(x, null); 			//update counts
						PreferredClusterVersions.put(x, null); 	//update sequence versions
					}
				}
			}

		}
		
//		//debugging: find clusters with min larger than 1.
//		for (Integer x : AllMinClusterCounts.keySet()){
//			if (AllMinClusterCounts.get(x) != null){
//				if (AllMinClusterCounts.get(x) > 1){
//					System.out.println("Min is larger than 1: " + x);
//				}
//			}
//		}
		
		// (2) For cases with multiple paralogs to choose from in alignment, pick best (length similarity)
		// 	   and write to file
		
		try {
			
			//buffered writer
			BufferedWriter bw = new BufferedWriter(new FileWriter(FileName));
						
			//check all organisms
			for (String s : GangFeaturesHash.keySet()){
							
				//retrieve starting list of features
				LinkedHashMap<Integer,LinkedList<GenomicFeature>> FeatureHash = GangFeaturesHash.get(s);
				
				//Initialize a protein sequence + header
				String str = "";
				String strHeader = ">" + s + "\n";
				
				int GFCounter = 0;
				String ClustCheck = "";
				
				//Initialize a list of the genes
				LinkedList<GenomicFeature> SortedGenes = new LinkedList<GenomicFeature>();
				
				//pick appropriate sequences
				for (Integer x : FeatureHash.keySet()){
					
					//Initialize a new re-ordering
					LinkedList<GenomicFeature> FeaturesInOrder = new LinkedList<GenomicFeature>();
					
					//retrieve all feature instances of this particular cluster, in this particular organism
					LinkedList<GenomicFeature> ClustFeatures = FeatureHash.get(x);
					
					int RefSize = PreferredClusterVersions.get(x).get(0).Size;
					
					//sort list by size diff
					for (int i = 0; i < ClustFeatures.size()-1; i++){
						for (int j = 0; j <ClustFeatures.size() -1; j++){
							if (Math.abs(ClustFeatures.get(i).Size - RefSize) <
									Math.abs(ClustFeatures.get(j).Size - RefSize)){
								GenomicFeature Temp = ClustFeatures.get(i);
								ClustFeatures.set(i, ClustFeatures.get(j));
								ClustFeatures.set(j, Temp);
							}
						}
					}
					
//					//debugging - take head of list
//					if (ClustFeatures.size() > 1){
//						for (int i = 0; i < ClustFeatures.size(); i++){
//							System.out.println(ClustFeatures.get(i).Size);
//						}
//						System.out.println("Breakpoint!");
//					}

					//if they match in size, no need for modifications
					if (AllMinClusterCounts.get(x) == ClustFeatures.size()){
						FeaturesInOrder = ClustFeatures;
					} else {
						FeaturesInOrder.add(ClustFeatures.get(0));
					}
					
					//sort sub-list + add
					Collections.sort(FeaturesInOrder, new ClusterNumSorter());
					SortedGenes.addAll(FeaturesInOrder);
					
				}
				
				//sorted list
				Collections.sort(SortedGenes, new ClusterNumSorter());
				
				//retrieve protein sequences from sorted list
				for (GenomicFeature GF : SortedGenes){
					
					//For printing to file
					GFCounter++;
					ClustCheck = ClustCheck + "_" + Integer.valueOf(GF.ClusterNum);
					str = str + GF.ProteinSequence;
					
					//output
					if (FinalClusters.get(GF.ClusterNum) != null){
						LinkedList<GenomicFeature> Features2Add = FinalClusters.get(GF.ClusterNum);
						Features2Add.add(GF);
						FinalClusters.put(GF.ClusterNum, Features2Add);
					} else {
						LinkedList<GenomicFeature> Features2Add = new LinkedList<GenomicFeature>();
						Features2Add.add(GF);
						FinalClusters.put(GF.ClusterNum, Features2Add);
					}
				}

//				//debugging - check size + gene order of protein sequences.
//				System.out.println("Gang " + GGNumber
//						+ ", Org:" + s + ", Genes: " + GFCounter 
//						+ " " + ClustCheck);
				
				//once genes written, write to file
				str = str + "\n\n";
				
				//write to file
				bw.write(strHeader);
				bw.write(str);
				
			}
			
			//close file stream
			bw.close();
			
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	
		
		//return statement
		return FinalClusters;
		
	}
}
