package Spring2014;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.LinkedHashMap;
import java.util.LinkedList;

public class SizeRangeStats {

	/**
	 * @param args
	 */
	
	//Fields
	public static final String ContextSetFile = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/ContextSets/7K_0.829.txt";
	public static LinkedHashMap<String,LinkedHashMap<Integer,LinkedList<Integer>>> GangData;
	public static LinkedHashMap<Integer,Integer> PBCV1_Mapping; //<K,V> = <wrong, correct>
	
	//manually map gang to gang
	public static void BuildPBCV1GangNumbering(){
		
		//initialize
		PBCV1_Mapping = new LinkedHashMap<Integer,Integer>();
		
		//								 <old mapping, correct mapping>
		
		//build mapping
		PBCV1_Mapping.put(8,1);
		PBCV1_Mapping.put(5,2);
		PBCV1_Mapping.put(9,3);
		PBCV1_Mapping.put(1,4);
		PBCV1_Mapping.put(2,5);
		PBCV1_Mapping.put(12,6);
		PBCV1_Mapping.put(20,7);
		PBCV1_Mapping.put(18,8);
		PBCV1_Mapping.put(19,9);
		PBCV1_Mapping.put(15,10);
		PBCV1_Mapping.put(24,11);
		PBCV1_Mapping.put(28,12);
		PBCV1_Mapping.put(3,13);
		PBCV1_Mapping.put(4,14);
		PBCV1_Mapping.put(11,15);
		PBCV1_Mapping.put(14,16);
		PBCV1_Mapping.put(22,17);
		PBCV1_Mapping.put(27,18);
		PBCV1_Mapping.put(6,19);
		PBCV1_Mapping.put(7,20);
		PBCV1_Mapping.put(17,21);
		PBCV1_Mapping.put(21,22);
		PBCV1_Mapping.put(23,23);
		PBCV1_Mapping.put(26,24);
		PBCV1_Mapping.put(31,25);
		
	}
	
	//Import context set data
	public static void ImportContextSetData(){
		
		//Initialize output structure
		GangData = new LinkedHashMap<String,LinkedHashMap<Integer,LinkedList<Integer>>>();
		
		//read in data from file into output structure
		try {
			//open file stream
			BufferedReader br = new BufferedReader(new FileReader(ContextSetFile));
			String Line = null;
			
			//parse file
			while ((Line = br.readLine()) != null){
				
				//split string by tabs
				String[] L = Line.split("\t");
				
				//data from string
				String Org = L[0];
				int GeneStart = Integer.parseInt(L[2]);
				int GeneStop = Integer.parseInt(L[3]);
				int GangNum = Integer.parseInt(L[4]);
				
				//retrieve org hash - initialize if necessary
				LinkedHashMap<Integer, LinkedList<Integer>> GangBoundaries;
				if (GangData.get(Org) != null){
					GangBoundaries = GangData.get(Org);
				} else{
					GangBoundaries = new LinkedHashMap<Integer, LinkedList<Integer>>();
				}
				
				//retrieve boundary list - initialize if necessary
				LinkedList<Integer> Boundaries;
				if (GangBoundaries.get(GangNum) != null){
					Boundaries = GangBoundaries.get(GangNum);
				} else{
					Boundaries = new LinkedList<Integer>();
				}
				
				//adjust boundary list
				if (Boundaries.size() == 0){
					Boundaries.add(GeneStart);
					Boundaries.add(GeneStop);
				} else {
					Boundaries.removeLast();
					Boundaries.add(GeneStop);
				}
				
				//update data storage
				GangBoundaries.put(GangNum, Boundaries);
				GangData.put(Org,GangBoundaries);
				
			}
			
			//close file stream
			br.close();
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//perform all imports
	public static void Initializations(){
		ImportContextSetData();
		BuildPBCV1GangNumbering();
	}
	
	//Export data table - PBCV1 data
	public static void ExportPBCV1DataTable(String OutputFile){
		
		try {
			
			//open file stream
			BufferedWriter bw = new BufferedWriter(new FileWriter(OutputFile));
			
			//write header
			String Header = "#Gang\tStart\tStop\n";
			bw.write(Header);
			bw.flush();
			
			LinkedHashMap<Integer,LinkedList<Integer>> PBCV1Data = GangData.get("PBCV1");
			
			//make hash for export
			LinkedHashMap<Integer,String> DataHash = new LinkedHashMap<Integer,String>();
			
			//retrieve data from hash
			for (Integer x : PBCV1Data.keySet()){
				
				//retrieve range
				LinkedList<Integer> Range = PBCV1Data.get(x);
				
				//build line
				String Line = String.valueOf(PBCV1_Mapping.get(x)) + "\t"
						+ String.valueOf(Range.get(0)) + "\t"
						+ String.valueOf(Range.get(1)) + "\n";
				
				//write to structure for later export
				DataHash.put(PBCV1_Mapping.get(x),Line);
			}
			
			//write data to file
			for (int i = 1; i <= 25; i++){
				String Line = DataHash.get(i);
				bw.write(Line);
				bw.flush();
			}

			//close file stream
			bw.close();
			
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	//Export data table - PBCV1 data
	public static void ExportSummaryDataTable(String OutputFile){
		
		try {
			
			//open file stream
			BufferedWriter bw = new BufferedWriter(new FileWriter(OutputFile));
			
			//write header
			String Header = "#Gang\tMin_size\tMax_size\tDifference\tRatio\n";
			bw.write(Header);
			bw.flush();
			
			//create hash for max and min
			LinkedHashMap<Integer, Integer> MaxMapping = new LinkedHashMap<Integer, Integer>();
			LinkedHashMap<Integer, Integer> MinMapping = new LinkedHashMap<Integer, Integer>();
			
			//iterate through each species
			for (String s : GangData.keySet()){
				
				//retrieve data
				LinkedHashMap<Integer, LinkedList<Integer>> Data
					= GangData.get(s);
				
				//iterate through all gangs
				for (Integer x : Data.keySet()){
					
					//retrieve range
					LinkedList<Integer> Range = Data.get(x);
					
					//determine size
					int RangeSize = Range.get(1) - Range.get(0);
					
					//not all gang numbers have a mapping
					if (PBCV1_Mapping.get(x) != null){
						
						// corrected gang number
						int GangNum = PBCV1_Mapping.get(x);
						
						//check exising hash for data, update size - max
						if (MaxMapping.get(GangNum) != null){
							
							//retrieve old size
							int Size = MaxMapping.get(GangNum);
							
							//if this data point is a new maximum, update
							if (Size < RangeSize){
								MaxMapping.put(GangNum, RangeSize);
							}
						} else{
							MaxMapping.put(GangNum, RangeSize);
						}
						
						//check exising hash for data, update size - min
						if (MinMapping.get(GangNum) != null){
							
							//retrieve old size
							int Size = MinMapping.get(GangNum);
							
							//if this data point is a new maximum, update
							if (Size > RangeSize){
								MinMapping.put(GangNum, RangeSize);
							}
						} else{
							MinMapping.put(GangNum, RangeSize);
						}
					}
					

					
				}
			}
		
			//write all data to file
			for (int i = 1; i <= 25; i++){
				
				//retrieve values, compute ratio
				Double ratio = (double) Math.round(100*(double) MaxMapping.get(i) / (double) MinMapping.get(i)) / 100;
				
				//write line from data
				String Line = String.valueOf(i) + "\t"
						+ String.valueOf(MinMapping.get(i)) + "\t"
						+ String.valueOf(MaxMapping.get(i)) + "\t"
						+ String.valueOf(MaxMapping.get(i)-MinMapping.get(i)) + "\t"
						+ String.valueOf(ratio) + "\n";
				
				//write line to file
				bw.write(Line);
				bw.flush();
				
			}
			
			//close file stream
			bw.close();
			
		} catch (Exception ex){
			ex.printStackTrace();
		}
	}
	
	
	// main method
	public static void main(String[] args) {
		
		//Import gang data, map numbering
		Initializations();
		
		//Export PBCV1 data
		String PBCV1Ranges = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/PBCV1_Ranges.txt";
		ExportPBCV1DataTable(PBCV1Ranges);
		
		//Export summary of gang sizes
		String GangSizeSummary = "/Users/phillipseitzer/Dropbox/GeneGangsPaper/California/Phil/MultipleMemberGeneGangs/GeneGangs/GangSizeSummary.txt";
		ExportSummaryDataTable(GangSizeSummary);
		
		//final output message
		System.out.println("All Processes Successfully Completed!");
	}

}
